package easik;

import javax.swing.UIManager;
import easik.ui.ApplicationFrame;
import easik.ui.Help;

import java.awt.image.*;

/** 
 * Easik is the starting point of the application. It creates an instance of the Application
 * frame and then allows for user interaction.
 */

public class Easik {

	/** The instance of this sketch */
	private static Easik _instance;
	
	/** The current settings of this application, as defined in the ~/.easik.properties file */
	private EasikSettings _settings;
	
	/** The current frame */
	private ApplicationFrame _theFrame;

	/** The help files */
	private Help _bigHelp;

	/** The current version of EASIK. */
	public static final String VERSION = "2.0";

	/**
	 * The current svn revision of Easik.java.  This is mainly for internal use, though
	 * it also shows up as a build number in "About Easik".
	 */
	public static final int REVISION = Integer.parseInt("$Revision: 756 $".replaceAll(".*?(\\d+).*", "$1"));

	/**
	 * Entry point of the Application. Creates an instance of the application.
	 * 
	 * @param args	Program parameters; currently ignored
	 */
	public static void main(final String[] args) {
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				createAndShowGUI(args);
			}
		});
	}

	public static void createAndShowGUI(String[] args) {

		// On a mac, set the application to use the normal menus at the top of the screen
		// This has to be *first*, before other AWT/Swing items load up
		if (EasikConstants.RUNNING_ON_MAC) {
			System.setProperty("apple.laf.useScreenMenuBar", "true");
		}


		// Use the system look and feel, because the Java look and feel, while consistent on every
		// platform, is consistently *wrong* on every platform.
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
			System.err.println("Unable to set system look and feel: " + e.getMessage());
		}

		Easik e = Easik.getInstance();

		// If a filename is specified, try to load it as an overview.
		if (args.length > 0) {
			java.io.File f = new java.io.File(args[0]);
			e.getFrame().getOverview().openOverview(f);
		}
	}
	
	/**
	 * Creates an instance of the application frame and then sets it to be visible.
	 * <p>
	 * Creating an instance of this class will start running an instance of the application. 
	 */
	private Easik() {
		_instance = this;
		_settings = new EasikSettings();
		_theFrame= new ApplicationFrame();

		if (EasikConstants.RUNNING_ON_MAC) {
			try {
				getClass().getClassLoader().loadClass("easik.os.OSX").newInstance();
			}
			catch (Exception e) {} // All sorts of exceptions can be fired from the above; just ignore them.
		}

		_theFrame.setVisible(true);
	}

	/**
	 * Singleton method to get the one instance
	 * 
	 * @return The one instance.
	 */
	public static Easik getInstance() {
		if (_instance == null) {			
			new Easik();
		} 		
		return _instance;		
	}	
	
	/**
	 * Gets the help pane associated with this instance.
	 * 
	 * @return The Easik help pane
	 */
	public Help getHelp() {
		return _bigHelp == null ? _bigHelp = new Help() : _bigHelp;
	}
	
	/**
	 * Returns the current application frame
	 * 
	 * @return The frame
	 */
	public ApplicationFrame getFrame() {
		return _theFrame;
	}

	/**
	 * Returns the EasikSettings containing the current settings
	 * 
	 * @return The easik settings
	 * @since 2008-05-08
	 */
	public EasikSettings getSettings(){
		return _settings;
	}
}
