package easik.database.driver;

import easik.database.DatabaseDriver;
import easik.database.types.*;
import easik.sketch.Sketch;

import java.sql.*;


import java.util.Arrays;
import java.util.Properties;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

/**
 * EASIK MySQL database driver.  Options supported by this driver are as follows.  Note that almost
 * all of these options are optional, however getting a Connection object (from getConnection()) will
 * fail if certain options aren't provided.  This is used, for example, when a MySQL driver is needed
 * but a connection is not: for example, when producing an SQL dump file.  Note that all of the options
 * marked as required for getConnection() are also required when exporting to a database (as
 * the exporter calls getConnection()).
 *
 * <ul>
 * 	<li><code>database</code> &mdash; required; the name of the database to use <i>[required for getConnection() and getSketchExporter()]</i></li>
 * 	<li><code>hostname</code> &mdash; the hostname of the server to connect to <i>[optional; defaults to localhost]</i></li>
 * 	<li><code>port</code> &mdash; the port of the server to connect to <i>[optional; if not specified the default will be used]</i></li>
 * 	<li><code>username</code> &mdash; the username to use when connecting <i>[required for getConnection()]</i></li>
 * 	<li><code>password</code> &mdash; the password to use when connecting <i>[required for getConnection()]</i></li>
 * 	<li><code>connectProperties</code> &mdash; can be set to a {@link java.util.Properties} object
 * 		containing MySQL driver-specific options <i>[optional; only has an effect when calling getConnection()]</i><code></li>
 * 	<li><code>quoteIdentifiers</code> &mdash; if set to the string <code>"true"</code>, all
 * 		identifiers (tables, columns, etc.) will be quoted.  This means column and table names may
 * 		contain any characters (including spaces).  Note that certain reserved names (e.g.
 * 		"Table"&mdash;see the list below) will always be quoted.  Also note that if quoteIdentifiers
 * 		is *not* enabled, sequences of characters other than a-z, A-Z, 0-9, and _ will be converted to
 * 		underscores.</li>
 * </ul>
 *
 * @see easik.database.DatabaseDriver
 */
public class MySQL extends DatabaseDriver {
	/**
	 * Reserved identifier names in MySQL.  If we encounter a table/column
	 * using one of these names, we force it to be quoted.
	 */
	public static final Set<String> RESERVED_KEYWORDS = new HashSet<String>(Arrays.asList((
			"ADD ALL ALTER ANALYZE AND AS ASC ASENSITIVE "+
			"BEFORE BETWEEN BIGINT BINARY BLOB BOTH BY "+
			"CALL CASCADE CASE CHANGE CHAR CHARACTER CHECK COLLATE COLUMN CONDITION CONSTRAINT "+
			"CONTINUE CONVERT CREATE CROSS CURRENT_DATE CURRENT_TIME CURRENT_TIMESTAMP CURRENT_USER CURSOR "+
			"DATABASE DATABASES DAY_HOUR DAY_MICROSECOND DAY_MINUTE DAY_SECOND DEC DECIMAL DECLARE DEFAULT "+

			"DELAYED DELETE DESC DESCRIBE DETERMINISTIC "+
			"DISTINCT DISTINCTROW DIV DOUBLE DROP DUAL "+ 
			"EACH ELSE ELSEIF ENCLOSED ESCAPED EXISTS EXIT EXPLAIN "+
			"FALSE FETCH FLOAT FLOAT4 FLOAT8 FOR FORCE FOREIGN FROM FULLTEXT "+
			"GRANT GROUP HAVING	HIGH_PRIORITY HOUR_MICROSECOND HOUR_MINUTE HOUR_SECOND "+ 
			"IF IGNORE IN INDEX INFILE INNER INOUT INSENSITIVE INSERT "+ 
			"INT INT1 INT2 INT3 INT4 INT8 INTEGER INTERVAL INTO IS ITERATE "+ 
			"JOIN KEY KEYS KILL "+ 
			"LEADING LEAVE LEFT LIKE LIMIT LINES LOAD LOCALTIME	LOCALTIMESTAMP "+ 
			"LOCK LONG LONGBLOB LONGTEXT LOOP LOW_PRIORITY "+
			"MATCH MEDIUMBLOB MEDIUMINT MEDIUMTEXT MIDDLEINT MINUTE_MICROSECOND MINUTE_SECOND MOD MODIFIES "+ 
			"NATURAL NOT NO_WRITE_TO_BINLOG NULL NUMERIC "+ 
			"ON OPTIMIZE OPTION OPTIONALLY OR ORDER OUT OUTER OUTFILE "+ 
			"PRECISION PRIMARY PROCEDURE PURGE "+ 
			"READ READS REAL REFERENCES REGEXP RELEASE RENAME REPEAT REPLACE REQUIRE "+
			"RESTRICT RETURN REVOKE RIGHT RLIKE "+ 
			"SCHEMA SCHEMAS SECOND_MICROSECOND SELECT SENSITIVE SEPARATOR SET SHOW SMALLINT SONAME "+ 
			"SPATIAL SPECIFIC SQL SQLEXCEPTION SQLSTATE SQLWARNING SQL_BIG_RESULT SQL_CALC_FOUND_ROWS "+ 
			"SQL_SMALL_RESULT SSL STARTING STRAIGHT_JOIN "+ 
			"TABLE TERMINATED THEN TINYBLOB TINYINT TINYTEXT TO TRAILING TRIGGER TRUE "+ 
			"UNDO UNION UNIQUE UNLOCK UNSIGNED UPDATE USAGE USE USING UTC_DATE UTC_TIME UTC_TIMESTAMP "+
			"VALUES VARBINARY VARCHAR VARCHARACTER VARYING "+ 
			"WHEN WHERE WHILE WITH WRITE "+ 
			"XOR YEAR_MONTH ZEROFILL ").split("\\s+")));

	/** Creates a new MySQL database driver object.  Should not be called
	 * directly.  @see DatabaseDriver.loadDriver(String, Map)
	 */
	public MySQL(Map <String, ?> opts) throws LoadException, SQLException{ 
		options = new HashMap<String, Object>(opts);
	}

	/** Returns true if enough options are specified to establish a connection.
	 * At a minimum, <code>database</code>, <code>username</code>, and
	 * <code>password</code> options are required, and <code>database</code>
	 * and <code>username</code> must be non-empty strings.
	 */
	public boolean isConnectable() {
		// database and username have to be non-empty; password just has to be
		// specified (but can be empty):
		return hasOption("database") && hasOption("username") && options.containsKey("password"); 
	} 

	/**
	 * Creates a database corresponding to the current set of options.
	 * @param drop Indicates if the database should be droped before creation, should a naming conflict arise.
	 * @return Success of creation.
	 */
	public boolean createDatabase(boolean drop) throws LoadException, SQLException {
		options.put("createDatabase", "true");
		Connection conn = getConnection();

		if(conn == null) return false;

		if(drop)
			conn.createStatement().executeUpdate("DROP DATABASE IF EXISTS " + options.get("database"));

		conn.createStatement().executeUpdate("CREATE DATABASE " + options.get("database"));

		sqlConn.setCatalog((String)getOption("database"));

		return true;
	}

	public Connection getConnection() throws LoadException, SQLException {
		if (sqlConn != null) return sqlConn;

		if (!isConnectable())
			throw new LoadException("Insufficient driver parameters for establishing a database connection: at least database, username, and password are required");

		StringBuilder dsn = new StringBuilder("jdbc:mysql://");
		if (optionMatches("hostname", "^[A-Za-z0-9.-]+$"))
			dsn.append(options.get("hostname").toString());
		else
			dsn.append("localhost");

		// In theory, a port name can be used on some system (typically such systems have an /etc/services file),
		// which map a name like 'mysql' to a number like 3306; but we don't allow that here.
		if (optionMatches("port", "^[0-9]+$"))
			dsn.append(":" + options.get("port"));

		if( ! hasOption("createDatabase") ){
			dsn.append("/" + options.get("database"));
		}

		Properties props;
		if (options.containsKey("connectProperties") && options.get("connectProperties") instanceof Properties)
			props = new Properties((Properties) options.get("connectProperties"));
		else
			props = new Properties();

		// MySQL has a "false" default for the paranoid option which includes stack trace
		// information in any errors messages, so unless the DatabaseDriver creator explicitly
		// asks for it to be false, default it to true.
		if (!props.containsKey("paranoid"))
			props.setProperty("paranoid", "true");

		if (options.containsKey("username"))
			props.setProperty("user", options.get("username").toString());
		if (options.containsKey("password"))
			props.setProperty("password", options.get("password").toString());

		try {
			// Loading the MySQL Driver class registers it with the DriverManager
			Class.forName("com.mysql.jdbc.Driver");
		} catch (ClassNotFoundException cnfe) {
			throw new LoadException("Unable to load MySQL driver", cnfe);
		}

		sqlConn = DriverManager.getConnection(dsn.toString(), props);

		return sqlConn;
	}

	/**
	 * Creates a new SketchExporter object for this MySQL driver.  The database parameter
	 * must have been specified when this database driver was created.
	 *
	 * @throws LoadException if the <code>database</code> option was not
	 * 		specified when this MySQL database driver was created.
	 */
	public easik.database.export.MySQL getSketchExporter(Sketch sketch, Map<String, ?> exportOpts)
				throws LoadException {

		if (!hasOption("database")){
			throw new LoadException("MySQL exporter requires \"database\" parameter!");
		}

		return new easik.database.export.MySQL(sketch, this, exportOpts);
	}

	/**
	 * MySQL-specific identifier cleaning.  This varies depending on quoteIdentifiers: if enabled,
	 * we trim trailing spaces, and convert /, \, and . to underscores, and double-up ` characters.
	 * Otherwise, we fall back to SketchExporters cleaning.
	 *
	 * This still isn't perfect, however: MySQL's retarded design means that, prior to MySQL 5.1.6,
	 * databases and tables, even when quoted, must not contain anything not representable on disk
	 * (because of the terrible design decision that database and table names should be mapped
	 * directly to filenames).  In effect, this means a MySQL server running on a Windows system,
	 * for example, has entirely different naming constraints than a MySQL server running on a unix
	 * system.  We actually ignore this constraint, which means if we have entities/attributes
	 * with ".", "/", or "\" characters (or various characters not allowed in Windows filenames),
	 * the SQL won't work on MySQL servers older than 5.1.6.
	 */
	public String cleanId(Object id) {
		if (optionEnabled("quoteIdentifiers")){
			return id.toString().replaceAll("`", "``");
		}
		else
			return super.cleanId(id);
	}

	/**
	 * MySQL-specific identifier quoting.  If the "quoteIdentifiers" option is enabled, or the
	 * identifier is a reserved keyword, we quote identifiers using the MySQL-specific
	 * <code>`</code> character (as opposed to the SQL-standard <code>"</code> character).
	 */
	public String quoteId(Object id) {
		String cleaned = cleanId(id);
		if (optionEnabled("quoteIdentifiers") || RESERVED_KEYWORDS.contains(cleaned.toUpperCase()))
			return "`" + cleaned + "`";
		else
			return cleaned;
	}

	public String getTypeString(Type type) {
		if (type instanceof easik.database.types.Float)
			return "REAL"; // MySQL treats FLOAT as DOUBLE PRECISION
		else if (type instanceof Text)
			return "LONGTEXT"; // MySQL provides TINYTEXT, TEXT, MEDIUMTEXT, and LONGTEXT for no good reason.  Just use LONGTEXT.
		else if (type instanceof easik.database.types.Blob)
			return "LONGBLOB"; // Likewise, MySQL pointlessly has 4 BLOB types.
		else if (type instanceof easik.database.types.Timestamp)
			return "DATETIME"; // MySQL TIMESTAMPs are horrid auto-updating things
		else
			return type.toString();
	}
}
