package easik.database.export;

import easik.EasikConstants;
import easik.overview.vertex.*;
import easik.sketch.Sketch;
import easik.sketch.vertex.*;
import easik.sketch.edge.*;
import easik.sketch.keys.UniqueIndexable;
import easik.sketch.path.SketchPath;
import easik.sketch.constraint.*;
import easik.database.ColumnNaming;
import easik.database.Options;
import easik.database.DatabaseDriver;
import easik.view.vertex.*;

import java.util.regex.Pattern;
import java.io.*;
import java.util.*;
import java.text.SimpleDateFormat;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

/**
 * Base class for exporting a sketch to a file, string, or database connection.
 *
 * SketchExporter objects are not created directly, but rather by first getting
 * a DatabaseDriver object and calling getSketchExporter(Sketch, options) on it.
 *
 * In addition to per-driver options, the following options are recognized by all
 * drivers (though for some drivers, may have no effect).
 *
 * <ul>
 * 	<li><code>bigKeys</code> &mdash; if set to the string <code>"true"</code>, the generated tables
 * 		will use BIGINTs (or equivelant) instead of INTEGERs for primary keys and foreign key columns.
 * 		Use this if your table could conceivably contain more than 2 billion rows in its lifetime.</li>
 * </ul>
 */
public abstract class SketchExporter extends Options {
	protected Sketch sketch;

	protected ColumnNaming colNaming;

	protected Pattern startOfLine = Pattern.compile("^", Pattern.MULTILINE);

	// Records unique keys that get auto-created to enforce constraints during export
	protected UniqueHM addedUnique = new UniqueHM();

	protected class UniqueHM extends HashMap<EntityNode, LinkedList<HashSet<SketchEdge>>> {
		/* Returns true if the EntityNode plus uniques auto-generated by the
		 * export already enforce uniqueness on the provided edges
		 */
		public boolean isUnique(EntityNode node, SketchEdge... edges) {
			// If any of the edges is an injective edge, it's unique (so anything containing it is, too)
			for (SketchEdge e: edges) {
				if (e instanceof InjectiveEdge)
					return true;
			}

			// All SketchEdges other than InjectiveEdges are unique-indexable:
			HashSet<UniqueIndexable> test = new HashSet<UniqueIndexable>();
			for (SketchEdge e: edges) {
				if (e instanceof UniqueIndexable)
					test.add((UniqueIndexable) e);
			}

			// If the node has a unique key enforcing uniqueness, we're good:
			if (node.uniqueKeyOn(test) != null)
				return true;

			// Otherwise go through all our added sets, and return true is any of them
			// is a subset of test
			LinkedList<HashSet<SketchEdge>> nodeUniques = get(node);
			if (nodeUniques != null) {
				for (HashSet<SketchEdge> unique: nodeUniques) {
					if (test.containsAll(unique))
						return true;
				}
			}

			return false;
		}
		/* Adds the specified SketchEdges as a new unique set associated with the EntityNode */
		public void add(EntityNode node, SketchEdge... edges) {
			LinkedList<HashSet<SketchEdge>> nodeUniques = get(node);
			if (nodeUniques == null)
				nodeUniques = new LinkedList<HashSet<SketchEdge>>();

			nodeUniques.add(new HashSet<SketchEdge>(Arrays.asList(edges)));
		}
	};

	/**
	 * The available export modes.  Mode.TEXT is for SQL that will be returned
	 * as a single string; Mode.FILE is for SQL that is being written to a
	 * file; Mode.DATABASE is for an export being sent directly to a database.
	 * Currently, the main different between them is that comments are
	 * suppressed from a database export.
	 */
	protected enum Mode { TEXT, FILE, DATABASE };
	/**
	 * The current export mode.
	 */
	protected Mode mode;

	protected boolean hadAutoCommit;

	/**
	 * The end-of-statement string, by default a ";".  This will be set to the
	 * empty string when exporting directly to a database since queries are
	 * sent individually.  Since ";" is only required in TEXT and FILE mode,
	 * not in DATABASE mode, <code>$</code> should always be appended to
	 * statements requiring a terminator.
	 */
	protected String $;

	/**
	 * The database driver.  We need this even when not connecting to a
	 * database for handling quoting, identifier cleaning, and column naming
	 * (via ColumnNaming).
	 */
	protected DatabaseDriver dbd;

	/**
	 * Protected constructor to be used by subclasses to set common fields.
	 */
	protected SketchExporter(Sketch sk, DatabaseDriver db, Map<String, ?> exportOpts) {
		dbd = db;
		sketch = sk;
		options = new HashMap<String, Object>(exportOpts);
		$ = dbd.getStatementSeparator();
		colNaming = new ColumnNaming(db);
	}

	/**
	 * Returns the sketch being exported.
	 */
	public Sketch getSketch() {
		return sketch;
	}

	/**
	 * Returns the initialization string, which is a comment containing the
	 * export date and Easik name and URL.
	 */
	protected List<String> initialize() {
		String now = (new SimpleDateFormat("d MMMM yyyy, h:mm:ss a z").format(new Date())).toString();
		return comment("Created " + now + "\nCreated using EASIK, " + EasikConstants.EASIK_URL);
	}

	/**
	 * Some things in initialize may be allowed to fail; if we get an exception, it
	 * gets passed here, and if this returns true, we ignore that exception and keep
	 * going.  See the postgresql driver for an example of this in action.
	 *
	 * @param e the SQLException that occurred 
	 * @return true, if the SQL exception should be ignored, false if it is a proper error.
	 */
	protected boolean ignoreError(SQLException e) {
		return false;
	}

	/**
	 * Takes a string, returns either a single-element list containing the
	 * string formatted as a comment, or, if comments are suppressed (for
	 * instance, for a database export), an empty list.  By default, commenting
	 * involves adding "-- " to the beginning of each line, but individual
	 * drivers may do something else.
	 */
	public List<String> comment(String text) {
		List<String> ret;
		if (mode == Mode.DATABASE)
			ret = Collections.emptyList();
		else
			ret = Collections.singletonList(startOfLine.matcher(text).replaceAll("-- "));
		return ret;
	}

	/**
	 * Writes the database export SQL script to a file.
	 *
	 * @param outFile the file to write to
	 */
	public void exportToFile(File outFile) throws IOException {
		PrintStream out = new PrintStream(new BufferedOutputStream(new FileOutputStream(outFile)));

		mode = Mode.FILE;
		$ = dbd.getStatementSeparator();

		// Writes the initial comments, database name, and so on
		for (String p: initialize())
			out.println(p);

		// Now write the table (include foreign key references)
		for (String t: createTables())
			out.println(t);

		// Generate any necessary constraints
		for (String c: createConstraints())
			out.println(c);

		// Generate any additional extras needed by the database
		for (String x: createExtras())
			out.println(x);

		// Generare all views on this sketch
		for (String v : createViews())
			out.println(v);

		out.close();
	}

	/**
	 * Returns the SQL export as a string.
	 */
	public String exportToString() {
		mode = Mode.TEXT;
		$ = dbd.getStatementSeparator();

		StringBuilder sql = new StringBuilder();

		for (String p: initialize())
			sql.append(p+"\n");
		for (String t: createTables())
			sql.append(t+"\n");
		for (String c: createConstraints())
			sql.append(c+"\n");
		for (String x: createExtras())
			sql.append(x+"\n");
		for (String v: createViews())
			sql.append(v+"\n");

		return sql.toString();
	}

	/**
	 * Exports the database directly to a database.  The database driver this export
	 * object was created from must support getConnection().
	 *
	 * @throws DatabaseDriver.LoadException if the database driver cannot be loaded
	 * @throws SQLException if an SQL error occurs
	 */
	public void exportToDB() throws DatabaseDriver.LoadException, SQLException {
		mode = Mode.DATABASE;
		$ = ""; // We don't want statement terminators

		Connection conn = null;
		try {
			if(optionEnabled("createDatabase"))
				dbd.createDatabase(optionEnabled("dropDatabase"));
			if(optionEnabled("dropSchema"))
				dbd.dropAndRecreateSchema();

			conn = dbd.getConnection();

			preInit(conn);

			for (String statement: initialize()){
				// Some things in initialize may be allowed to fail; if we get an exception, pass it to ignoreError,
				// and if it returns true, just keep going.
				try { dbd.executeUpdate(statement); }
				catch (SQLException e) { if (!ignoreError(e)) throw e; }
			}
			postInit(conn);

			for (String statement: createTables())
				dbd.executeUpdate(statement);

			for (String statement: createConstraints())
				dbd.executeUpdate(statement);

			for (String statement: createExtras())
				dbd.executeUpdate(statement);

			for (String statement: createViews())
				dbd.executeUpdate(statement);

			finished(conn);
		}
		catch (SQLException sqle) {
			System.err.println("Caught exception: " + sqle.getMessage());
			// Catch the exception and rethrow it, but first call aborted() to abort the current transaction.
			// aborted() might throw another SQLException; if so we just ignore it.
			if (conn != null) {
				try {
					aborted(conn);
				} catch (SQLException e2) {}
			}

			throw sqle;
		}
	}

	/**
	 * Called immediately before initialize() when doing a database export.
	 * This turns on auto-commit (it'll be turned off after initialization, in
	 * postInit(), and saved the old auto-commit value to be restored after the
	 * export is complete.
	 */
	protected void preInit(Connection conn) throws SQLException {
		hadAutoCommit = conn.getAutoCommit();
		conn.setAutoCommit(true);
	}

	/**
	 * Called immediately after initialize() when doing a database export.
	 * The default operation turns off auto-commit mode (i.e. it begins a
	 * transaction).
	 */
	protected void postInit(Connection conn) throws SQLException {
		conn.setAutoCommit(false);
	}

	/**
	 * Called upon successful completion of the creation queries.  By default
	 * the transaction is committed, the original auto-commit setting of the
	 * connection is restored and we set a flag in our sketch indicating that 
	 * we've exported to a database.
	 */
	protected void finished(Connection conn) throws SQLException {
		conn.commit();
		conn.setAutoCommit(hadAutoCommit);
		sketch.setSynced(true);
		sketch.setDirty();
	}
	/**
	 * Called if an SQLException occurs during database creation.  The default
	 * operation is to rollback the current transaction, then restore the original
	 * auto-commit setting.
	 */
	protected void aborted(Connection conn) throws SQLException {
		conn.rollback();
		conn.setAutoCommit(hadAutoCommit);
	}

	/**
	 * Returns the list of queries that creates the needed tables, including
	 * foreign key references, but not including data constraints.  The default
	 * operation is to figure out an appropriate creation order and then call
	 * createTable(EntityNode, boolean) for each table.
	 *
	 * @return list of queries to run to create the tables
	 */
	protected List<String> createTables() {
		LinkedList<String> sql = new LinkedList<String>();

		// Get the entities of the sketch; as we add them to the output string, we remove them
		// from here and add to 'created'.  When entities is empty, we're done.
		HashSet<EntityNode> entities = new HashSet<EntityNode>(sketch.getEntities());

		// As we create entities, we add them to this HashSet
		HashSet<EntityNode> created = new HashSet<EntityNode>();

		createTables(sql, entities, created);

		return sql;
	}

	/**
	 * Recursive method to create table definitions.  Each iteration exhaustively determines
	 * and generates all unconnected EntityNodes, then, if any EntityNodes remain, arbitrarily
	 * removes one (deferring creation of its references), and recurses, until all tables
	 * have been created.
	 */
	private void createTables(
			List<String> tables,
			HashSet<EntityNode> entities,
			HashSet<EntityNode> created) {

		NON_EMPTY: while (!entities.isEmpty()) {
			// Start out looking for entities with no edges to so-far uncreated entities
			ENTITIES: for (EntityNode n: entities) {
				for (SketchEdge e: n.getOutgoingEdges()) {
					EntityNode target = e.getTargetEntity();
					// if an EntityNode target isn't already created, skip this vertex (for now)
					if (!created.contains(target))
						continue ENTITIES;
				}
				// Also check shadow edges:
				for (SketchEdge e: n.getShadowEdges()) {
					EntityNode target = e.getTargetEntity();
					// if an EntityNode target isn't already created, skip this vertex (for now)
					if (!created.contains(target))
						continue ENTITIES;
				}

				// If we got here, all the entity's targets exist, so we can go ahead
				// with the table creation.
				tables.addAll(createTable(n, true));
				entities.remove(n);
				created.add(n);
				continue NON_EMPTY; // Restart the non-empty loop, since we've changed entities.
			}

			// If we get here, entities still exist, but they all point to entities that
			// don't yet exist--this can happen, for instance, with: A -> B and B -> A,
			// or (even simpler), a self-reference (A -> A).
			break NON_EMPTY;
		}

		if (!entities.isEmpty()) {
			EntityNode someNode = entities.iterator().next();
			tables.addAll(createTable(someNode, false));
			entities.remove(someNode);
			created.add(someNode);

			// Recurse, creating the rest of the tables
			createTables(tables, entities, created);

			// When the remaining tables are created, we are able to create the reference:
			tables.addAll(createReferences(someNode));
		}
	}

	/**
	 * Returns the table creation statement.  If includeReferences is true, the table
	 * output should include reference (foreign key) information; otherwise, this
	 * information should be skipped (generally in such a case, createReferences will
	 * be called later to access that information.
	 *
	 * @param table the EntityNode of the table to create
	 * @param includeReferences a boolean indicating whether or not the generated table
	 * 		  				    scheme may include references.
	 * @return a string of SQL that generates the table
	 */
	protected abstract List<String> createTable(EntityNode table, boolean includeReferences);

	/**
	 * Returns a string creating this table's references to other tables.  Normally
	 * this method is only called when reference loops exist (since at least
	 * one of the references must be deferred), however implementing classes are, of
	 * course, free to call this method from createTable (if inline references aren't
	 * used/supported, for instance).
	 *
	 * @param table the EntityNode of the table whose outgoing references are to be created
	 * @return a string of SQL that generates the references
	 */
	protected abstract List<String> createReferences(EntityNode table);

	/**
	 * Generates database constraints from the sketch constraints.  This method
	 * is not generally overridden by drivers: the individual
	 * createConstraint(...) methods are called for each existing constraint.
	 *
	 * @return list of queries to create the constraints.
	 */
	protected List<String> createConstraints() {
		LinkedList<String> constraintSQL = new LinkedList<String>();
		List<Constraint> constraints = sketch.getConstraints();

		int id = 0;
		for (Constraint c: constraints) {
			if (c instanceof CommutativeDiagram)
				constraintSQL.addAll(createConstraint((CommutativeDiagram) c, id++ + ""));
			else if (c instanceof ProductConstraint)
				constraintSQL.addAll(createConstraint((ProductConstraint) c, id++ + ""));
			else if (c instanceof PullbackConstraint)
				constraintSQL.addAll(createConstraint((PullbackConstraint) c, id++ + ""));
			else if (c instanceof EqualizerConstraint)
				constraintSQL.addAll(createConstraint((EqualizerConstraint) c, id++ + ""));
			else if (c instanceof SumConstraint)
				constraintSQL.addAll(createConstraint((SumConstraint) c, id++ + ""));
			else
				System.err.println("Unknown constraint type encountered: " + c.getClass());
		}

		return constraintSQL;
	}

	/**
	 * Returns a list of queries to run to create database constraints
	 * enforcing the passed-in commutative diagram.
	 *
	 * @param constraint the commutative diagram
	 * @param id a unique ID (typically a number) that can be used in constraint names
	 * @return list of queries to create database enforcement of the constraint
	 */
	protected abstract List<String> createConstraint(CommutativeDiagram constraint, String id);
	/**
	 * Returns a list of queries to run to create database constraints
	 * enforcing the passed-in product constraint.
	 *
	 * @param constraint the product constraint
	 * @param id a unique ID (typically a number) that can be used in constraint names
	 * @return list of queries to create database enforcement of the constraint
	 */
	protected abstract List<String> createConstraint(ProductConstraint constraint, String id);
	/**
	 * Returns a list of queries to run to create database constraints
	 * enforcing the passed-in equalizer constraint.
	 *
	 * @param constraint the equalizer constraint
	 * @param id a unique ID (typically a number) that can be used in constraint names
	 * @return list of queries to create database enforcement of the constraint
	 */
	protected abstract List<String> createConstraint(EqualizerConstraint constraint, String id);
	/**
	 * Returns a list of queries to run to create database constraints
	 * enforcing the passed-in pullback.
	 *
	 * @param constraint the pullback constraint
	 * @param id a unique ID (typically a number) that can be used in constraint names
	 * @return list of queries to create database enforcement of the constraint
	 */
	protected abstract List<String> createConstraint(PullbackConstraint constraint, String id);
	/**
	 * Returns a list of queries to run to create database constraints
	 * enforcing the passed-in sum constraint.
	 *
	 * @param constraint the sum constraint
	 * @param id a unique ID (typically a number) that can be used in constraint names
	 * @return list of queries to create database enforcement of the constraint
	 */
	protected abstract List<String> createConstraint(SumConstraint constraint, String id);

	/**
	 * Used for databases to create any required extra output.  This might
	 * includes some sort of "fail" procedure, for example.  The default implementation
	 * adds nothing.
	 *
	 * @return a String containing the extra SQL
	 */
	protected List<String> createExtras() {
		return Collections.emptyList();
	}

	/**
	 * Generates views on the sketch.
	 * @return list of queries to create our views
	 */
	protected List<String> createViews(){
		LinkedList<String> viewSQL = new LinkedList<String>();
		HashSet<ViewNode> views = new HashSet(sketch.getViews());

		for(ViewNode node : views){
			viewSQL.addAll(createView(node));
		}

		return viewSQL;
	}

	/**
	 * Returns the list view creation statements that will create our view.
	 * @param node The view node containing the query nodes making up our view
	 * @return A list of of SQL Strings that will generate the view.
	 */
	protected abstract List<String> createView(ViewNode node);

	/**
	 * Returns an SQL query string that will generate the view represended by a given query node.
	 * @param node The query node containing the SELECT statements that will generate the view.
	 * @return And SQL query string that will generate the view represented by our node.
	 */
	protected abstract String createView(QueryNode node);

	

	/**
	 * Constructs a join clause for this path.  For instance,
	 * if the path is a -&gt; b -&gt; c -&gt; d, this would return
	 * <code>a JOIN b ON b.a_id = a.id JOIN c ON c.b_id = b.id JOIN d ON d.c_id = c.id</code>
	 * suitable for use in a select query.
	 *
	 * @param inEdges the path for which the selection is being made
	 * @param includeLast true if the last path target should be included.  If false, the
	 * 		last path will be omitted.
	 * @param initialOn if non-null, "ON " + initialOn will be added after the path source table name
	 * @return the string, suitable for use in a SELECT query.
	 */
	protected String joinPath(List<SketchEdge> inEdges, boolean includeLast, String initialOn) {
		LinkedList<SketchEdge> edges = new LinkedList<SketchEdge>(inEdges);
		StringBuilder joinClause = new StringBuilder(quoteId(edges.get(0).getSourceEntity().getName()));

		if (initialOn != null)
			joinClause.append(" ON " + initialOn);

		if (!includeLast && !edges.isEmpty()) edges.removeLast();

		for (SketchEdge e: edges) {
			EntityNode target = e.getTargetEntity();
			joinClause.append(" JOIN " + quoteId(target) + " ON " + qualifiedFK(e) + " = " + qualifiedPK(target));
		}

		return joinClause.toString();
	}
	/**
	 * Shortcut for calling <code>joinPath(edges, wantLast, null)</code>.
	 *
	 * @see joinPath(List, boolean, String)
	 */
	protected String joinPath(List<SketchEdge> edges, boolean includeLast) {
		return joinPath(edges, includeLast, null);
	}
	/**
	 * Shortcut method for calling <code>joinPath(path.getEdges(), wantLast)</code>
	 *
	 * @see joinPath(List, boolean)
	 */
	protected String joinPath(SketchPath path, boolean includeLast) {
		return joinPath(path.getEdges(), includeLast);
	}
	/**
	 * Shortcut method for calling <code>joinPath(path.getEdges(), true);</code>
	 *
	 * @see joinPath(List, boolean)
	 */
	protected String joinPath(List<SketchEdge> edges) {
		return joinPath(edges, true);
	}
	/**
	 * Shortcut method for calling <code>joinPath(path, true);</code>.
	 *
	 * @see joinPath(SketchPath, boolean)
	 */
	protected String joinPath(SketchPath path) {
		return joinPath(path.getEdges());
	}

	/**
	 * These are convenience wrappers around the same methods in
	 * easik.database.ColumnNaming.  See that class for documentation.
	 */
	public String tablePK(EntityNode table)		{ return colNaming.tablePK(table); }
	public String qualifiedPK(EntityNode table)	{ return colNaming.qualifiedPK(table); }
	public String tableFK(SketchEdge edge)		{ return colNaming.tableFK(edge); }
	public String qualifiedFK(SketchEdge edge)	{ return colNaming.qualifiedFK(edge); }

	/**
	 * Simple convenience wrappers around the same methods in the DatabaseDriver
	 */
	public String cleanId(Object identifier) { return dbd.cleanId(identifier); }
	public String quoteId(Object identifier) { return dbd.quoteId(identifier); }
}
