package easik.database.types;

import java.util.Map;
import java.util.Collections;
import java.sql.Types;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * A class representing a CHAR(n) SQL data type.
 * A CHAR differs from a VARCHAR primarily in that, in theory, a CHAR will always
 * occupy the full size (n), while a VARCHAR will only use as much storage space
 * as needed to store the inserted values.  Note that some databases (such as
 * MySQL) will silently convert CHAR columns to VARCHAR columns if the table
 * contains any other variable-size data type.  Also note that the maximum size
 * of the CHAR type is database-dependent; some databases in particular will not
 * allow more than 255.
 */
public class Char extends Type {
	private int size;

	/**
	 * Creates a new Char object.  The n specifies the required size of this
	 * char.  Where possible, the database should try to use a CHAR(n),
	 * but database limitations may convert this into other text types.
	 */
	public Char(int n) {
		size = n;
	}

	public int getSize() {
		return size;
	}
	public String toString() { return "CHAR("+size+")"; }

	public boolean verifyInput(String input){ return input.length() <= size; }

	/**
	 * Returns the attributes of this object
	 */
	public Map<String, String> attributes() {
		return Collections.singletonMap("size", size+"");
	}
	/**
	 * Recreates the object from the attributes returned by attributes().
	 */
	public Char(Map<String, String> attr) {
		this(Integer.parseInt(attr.get("size")));
	}

	public int getSqlType(){ return Types.CHAR; }

	public void bindValue(PreparedStatement ps, int col, String value) throws SQLException{
		ps.setString(col, value);
	}
}
