package easik.sketch;

import java.awt.geom.Rectangle2D;
import java.awt.Color;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.Image;
import java.awt.Point;
import java.io.File;
import java.util.*;

import java.sql.*;

import javax.swing.*;
import javax.swing.undo.AbstractUndoableEdit;

import org.jgraph.JGraph;
import org.jgraph.graph.*;
import org.jgraph.event.*;

import easik.*;
import easik.database.DatabaseDriver;
import easik.database.types.Type;
import easik.overview.Overview;
import easik.overview.vertex.*;
import easik.sketch.constraint.Constraint;
import easik.sketch.edge.GuideEdge;
import easik.sketch.edge.SketchEdge;
import easik.sketch.edge.SketchEdge.Cascade;
import easik.sketch.util.SketchFileIO;
import easik.sketch.util.graph.SketchGraphModel;
import easik.sketch.util.graph.SketchViewFactory;
import easik.sketch.vertex.EntityNode;
import easik.sketch.states.LoadingState;
import easik.sketch.states.BasicEditingState;
import easik.sketch.keys.UniqueIndexable;
import easik.sketch.keys.UniqueKey;
import easik.ui.*;
import easik.ui.datamanip.UpdateMonitor;
import easik.ui.tree.SketchInfoTreeUI;
import easik.ui.menu.popup.*;
import easik.view.View;
import easik.view.vertex.*;

/** 
 * A Sketch represents an EA Sketch diagram used to represent a database, this object
 * also extends the JGraph swing component, allowing it to be added directly to the
 * application's GUI.
 * 
 * When done with the current sketch, instead of creating a new one, the sketch should
 * simply be reinitialised, and it will become ready. Since a sketch is also a Swing component
 * it can get hairy creating a new one and changing all the references. This might have
 * become easier since the singletons were introduced, so feel free to try to change that.
 *  
 * @author Rob Fletcher 2005
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 */
public class Sketch extends JGraph {

	/** Hash Map of all entity nodes, indexed by their name */
	private LinkedHashMap<String, EntityNode> _entityNodes;

	/** Hash Map of all edges, indexed by their label */
	private LinkedHashMap<String, SketchEdge> _edges;

	/** Linked List of all constraints */
	private LinkedList<Constraint> _constraints;

	/** The current ApplicationFrame */
	private SketchFrame _sketchFrame;

	/** The current DocumentInfo */
	private DocumentInfo _docInfo;

	/** The State Manager monitoring this sketch */
	private SketchStateManager _stateManager;

	/** The overview in which this sketch exists */
	private Overview _theOverview;

	/** The current connection to a MySQL server */
	private DatabaseDriver _dbd;

	/** The connection parameters stored with the sketch, reflecting the last
	 * settings used with this sketch. */
	private HashMap<String, String> _connParams;

	/** An update monitor that can be used when performing operations on this sketch */
	private UpdateMonitor _updateMonitor;

	/** Map of views on this sketch indexed by name */
	private HashMap<String, ViewNode> _views;

	/** The default cascade modes for new edges of this sketch */
	private Cascade _defaultCascade, _defaultPartialCascade;

	/** Flag set if this sketch has been exported. If set, should lock editing */
	private boolean _syncLock;

	/** Keeps our model on hand */
	private SketchGraphModel model;

	/**
	 * The default constructor sets all the visual settings for the JGraph, as well as
	 * initialising the sketch to be empty. It also adds appropriate listeners for all 
	 * of the actions we are concerned with.
	 * 
	 * @param inFrame The application frame of the sketch
	 */
	public Sketch(SketchFrame inFrame, Overview inOverview) {
		super();

		setBackground(Easik.getInstance().getSettings().getColor("edit_canvas_background"));
		_sketchFrame = inFrame;
		_theOverview = inOverview;
		_syncLock = false;
		_stateManager = new SketchStateManager(this);
		setAntiAliased(true);
		setDisconnectable(false);
		setConnectable(false);
		setEditable(false);
		setSizeable(false);
		_docInfo = new DocumentInfo(_sketchFrame);
		initialiseSketch();
		getGraphLayoutCache().setAutoSizeOnValueChange(true);
		setSelectionModel(new DefaultGraphSelectionModel(this) {
			public Object[] getSelectables() {
				return Sketch.this.getStateManager().getSelectables();
			}
		});
		addGraphSelectionListener(new GraphSelectionListener() {
			public void valueChanged(GraphSelectionEvent gsEvent) {
				Sketch.this.getStateManager().selectionUpdated();
			}
		});

		// Set up mouse listener to watch for double clicks
		//  - Double clicks edit component
		addMouseListener(new MouseAdapter(){
			public void mouseClicked(MouseEvent e){
				// Ignore it if we're not in the basic editing state:
				if (!(Sketch.this.getStateManager().peekState() instanceof BasicEditingState)) {
					return;
				}

				boolean doubleClick = (e.getClickCount() > 1 && e.getButton() == MouseEvent.BUTTON1);
				Object[] selection = getSelectionCells();
				if (_sketchFrame.getMode() == SketchFrame.Mode.EDIT && doubleClick && selection.length > 0) {
					if( selection[0] instanceof EntityNode ){
						//hack to get rename action
						JMenuItem renameAction = new JMenuItem(new RenameInSketchAction(_sketchFrame));
						renameAction.doClick();
					}
					else if( selection[0] instanceof SketchEdge){
						//hack to get edit edge action
						JMenuItem editEdgeAction = new JMenuItem(new EditSketchEdgeAction(_sketchFrame));
						editEdgeAction.doClick();
					}
				}
			}
			public void mousePressed(MouseEvent e){ mouseClicked(e); }
		});

		String cascade = Easik.getInstance().getSettings().getProperty("sql_cascade", "restrict");
		_defaultCascade = cascade.equals("cascade") ? Cascade.CASCADE : Cascade.RESTRICT;
		String pCascade = Easik.getInstance().getSettings().getProperty("sql_cascade_partial", "set_null");
		_defaultPartialCascade = pCascade.equals("cascade") ? Cascade.CASCADE : pCascade.equals("restrict") ? Cascade.RESTRICT : Cascade.SET_NULL;

		updateUI();
	}

	/**
	 * Since this is a Swing component, this method is overloading a method of
	 * JGraph to adjust the look and feel. The feel we are changing is ignoring all but
	 * left clicks, allowing for right-click functionality not affecting the selections.
	 */
	public void updateUI() {
		setUI(new GraphUI());
		invalidate();
	}

	/**
	 *	Returns the parental sketch frame to whoever asks for it.
	 * @return The current sketch frame
	 */
	public SketchFrame getFrame() {
		return _sketchFrame;
	}

	/**
	 * Gets the document information
	 * @return The document information 
	 */
	public DocumentInfo getDocInfo(){
		return _docInfo;
	}

	/**
	 * When we initialise the sketch, we flush out all the data concerning the sketch
	 * itself.
	 *
	 *	This methods serves as a "new sketch" function. 
	 */
	public void initialiseSketch() {
		clearSelection();
		model = new SketchGraphModel(this);
		GraphLayoutCache glc = new GraphLayoutCache(model, new SketchViewFactory());
		setModel(model);
		setGraphLayoutCache(glc);

		_entityNodes = new LinkedHashMap<String, EntityNode>();
		_edges = new LinkedHashMap<String, SketchEdge>();
		_constraints = new LinkedList<Constraint>();
		if(_sketchFrame.getInfoTreeUI() != null){
			_sketchFrame.setInfoTreeUI(new SketchInfoTreeUI(_sketchFrame)); //Wipe Tree
			_sketchFrame.getInfoTreeUI().refreshTree();
		}
		_docInfo.reset();
		_views = new HashMap<String, ViewNode>();
		_connParams = new HashMap<String, String>();

		model.discardUndo();
	}

	/**
	 * Used to initialise a new sketch based on provided data (usually from the Sketch loading
	 * methods).
	 * 
	 * @param entityNodes A map of all of the entities in the sketch
	 * @param edges A map containing all of the edges in the sketch
	 * @param constraints A collection of the constraints of the sketch
	 * @param head The header created from the loaded XML file.
	 */
	public void initializeFromData(
		boolean syncLock,
		Map<String, EntityNode> entityNodes,
		Map<String, SketchEdge> edges,
		Collection<Constraint> constraints,
		DocumentInfo head,
		Map<String, String> connParams)
	{
		initialiseSketch();

		_syncLock = syncLock;
		_sketchFrame.setMode(syncLock ? SketchFrame.Mode.MANIPULATE : SketchFrame.Mode.EDIT);
		_constraints = new LinkedList<Constraint>();
		_docInfo = head;
		
		_sketchFrame.setTreeName(_docInfo.getName());

		//Add the entities, edges, and constraints:
		addEntity(entityNodes.values());
		addEdge(edges.values());
		for (Constraint c: constraints)
			addConstraint(c);

		setConnectionParams(connParams);

		refresh();

		model.discardUndo();
	}

	/**
	 * Overridden version of JGraph.getSelectionCells() that always returns an
	 * empty array (instead of null) for an empty selection.
	 */
	public Object[] getSelectionCells() {
		Object[] ret = super.getSelectionCells();
		if (ret == null)
			ret = new Object[0];
		return ret;
	}

	public SketchStateManager getStateManager(){
		return _stateManager;
	}

	/**
	 * Refreshes the specified nodes/edges in the sketch GUI.  If called with
	 * no arguments, all items are refreshed.
	 */
	public void refresh(GraphCell... cells) {
		setBackground(Easik.getInstance().getSettings().getColor("edit_canvas_background"));
		Object[] toRefresh = cells.length > 0 ? cells : getRoots();

		model.cellsChanged(toRefresh);

		if (cells.length == 0)
			super.refresh();
	}

	/**
	 * Overrides superclass's refresh() method to go through the above
	 */
	public void refresh() {
		refresh(new GraphCell[0]);
	}

	/**
	 * Returns an edge based on its name.
	 *
	 * @param name name of the edge desired
	 */
	public SketchEdge getEdge(String name) {
		return _edges.get(name);
	}

	/**
	 * Returns a collection of all of the edges in the sketch
	 *
	 * @return Collection of all the edges.
	 */
	public Collection<SketchEdge> getEdges() {
		return Collections.unmodifiableCollection(_edges.values());
	}

	/**
	 * Accessor for a single entity in the sketch.  Takes the entity name.
	 *
	 * @param entName the entity name
	 * @return EntityNode associated with that name, or null if the entity does not exist
	 */
	public EntityNode getEntity(String entName) {
		return _entityNodes.get(entName);
	}

	/**
	 * Accessor for the entities in the sketch
	 * @return Collection of the entities
	 */
	public Collection<EntityNode> getEntities() {
		return Collections.unmodifiableCollection(_entityNodes.values());
	}

	/**
	 * Accessor for the set of views on this sketch.
	 * @return Collection of the views on this sketch
	 */
	public Collection<ViewNode> getViews(){
		return Collections.unmodifiableCollection(_views.values());
	}

	/**
	 * Removes a view with a given name from our set of views.
	 * @param viewName The name of the view.
	 */
	public void removeView(String name){
		_views.remove(name);
	}

	/**
	 * Adds a view to our set of views
	 * @param view The view to add to our set
	 */
	public void addView(ViewNode view){
		_views.put(view.getName(), view);
	}

	/**
	 * Called internally by EntityNode when a node is renamed, to keep the
	 * entity node map consistent.  Should not be called directly; instead
	 * call sketch.getEntity("currentname").setName("newname").
	 *
	 * @see easik.sketch.vertex.EntityNode.setName(String)
	 * @param node the EntityNode being renamed
	 * @param oldName the old name of the node
	 * @param newName the candidate new name
	 * @return a string containing the final new node name, for EntityNode to use.
	 */
	public String nodeRenamed(EntityNode node, String oldName, String newName) {
		// If the name already exists, we have to rename it
		while (_entityNodes.containsKey(newName))
			newName = EasikTools.incrementName(newName);

		_entityNodes.put(newName, _entityNodes.remove(oldName));

		return newName;
	}

	/**
	 * Called internally by SketchEdge when an edge is renamed, to keep the
	 * edge map consistent.  Should not be called directly; instead just
	 * call edge.setName("newname").
	 *
	 * @see easik.sketch.edge.SketchEdge.setName(String)
	 * @param edge the edge being renamed
	 * @param oldName the old name of the edge
	 * @param newName the candidate new name
	 * @return a string containing the final new edge name, for SketchEdge to use.
	 */
	public String edgeRenamed(SketchEdge edge, String oldName, String newName) {
		// If the name already exists, we have to rename it
		while (_edges.containsKey(newName))
			newName = EasikTools.incrementName(newName);

		_edges.put(newName, _edges.remove(oldName));

		return newName;
	}

	/**
	 * Returns the list of constraints
	 * 
	 * @return The constraints
	 */
	public List<Constraint> getConstraints() {
		return Collections.unmodifiableList(_constraints);
	}

	/**
	 * Requests that an XML file be loaded into the sketch.
	 * 
	 * @param inputFile The file from which the data will be drawn.
	 * @return The success of the load
	 */
	public boolean loadFromXML(File inputFile) {
		if( ! SketchFileIO.graphicalSketchFromXML(inputFile, this)){
			_theOverview.removeSketch(_sketchFrame.getNode());
			return false;
		}
		else{
			return true;
		}
	}

	/**
	 * Removes an entity, and also cascades to remove all the arrows 
	 * involved with it.
	 * 
	 * @param toRemove The entity about to be removed
	 */
	public void removeEntity(final EntityNode toRemove) {

		//So we don't get a ConcurrentModificationException
		ArrayList<SketchEdge> removeEdges = new ArrayList<SketchEdge>();
		
		for (SketchEdge edge: _edges.values()) {
			if (edge.getSourceEntity().equals(toRemove) || edge.getTargetEntity().equals(toRemove)) {
				//add the edge to a list of edges to remove
				removeEdges.add(edge);
			}
		}

		model.beginUpdate();

		//Remove the edges
		for (SketchEdge e: removeEdges) {
			removeEdge(e);
		}
		_entityNodes.remove(toRemove.toString());
		getGraphLayoutCache().remove(new Object[] {toRemove});

		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_entityNodes.put(toRemove.toString(), toRemove);
			}
			public void redo() {
				super.redo();
				_entityNodes.remove(toRemove.toString());
			}
		});

		//Remove Entity from tree
		_sketchFrame.getInfoTreeUI().removeEntity(toRemove);

		model.endUpdate();
	}

	/**
	 * Removes an edge, also cascades to remove all constraints using it.
	 * 
	 * @param toRemove The edge about to be removed
	 */
	public void removeEdge(final SketchEdge toRemove) {
		model.beginUpdate();

		// Check for constraints that need these edges
		for (int i = 0; i < _constraints.size(); i++) {
			Constraint currentConstraint = (Constraint) _constraints.get(i);
			if (currentConstraint.hasEdge(toRemove)) {
				removeConstraint(currentConstraint);
				i--;
			}
		}

		if (toRemove instanceof UniqueIndexable) {
			EntityNode source = toRemove.getSourceEntity();

			boolean needCleanup = false;
			for (UniqueKey q: source.getUniqueKeys()) {
				if (q.removeElement((UniqueIndexable) toRemove))
					needCleanup = true;
			}

			if (needCleanup)
				UniqueKey.cleanup(source);
			// FIXME: undo: unique key cleanup stuff
		}

		_edges.remove(toRemove.getName());

		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_edges.put(toRemove.getName(), toRemove);
			}
			public void redo() {
				super.redo();
				_edges.remove(toRemove.getName());
			}
		});

		getGraphLayoutCache().remove(new Object[] {toRemove});

		model.endUpdate();
	}

	/**
	 * Removes a constraint and guide arrows
	 * 
	 * @param toRemove The constraint about to be removed
	 */
	public void removeConstraint(final Constraint toRemove) {
		model.beginUpdate();

		getGraphLayoutCache().remove(toRemove.getGuideEdges().toArray(new GuideEdge[0]));
		getGraphLayoutCache().remove(new Object[] {toRemove});

		final int position = _constraints.indexOf(toRemove);
		_constraints.remove(toRemove);

		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_constraints.add(position, toRemove);
			}
			public void redo() {
				super.redo();
				_constraints.remove(toRemove);
			}
		});

		//Remove Entity from tree
		_sketchFrame.getInfoTreeUI().removeConstraint(toRemove);

		model.endUpdate();
	}

	/**
	 * Saves the existing sketch as an XML file.
	 * 	
	 * @param outputFile The file to be written to
	 */
	public void saveToXML(File outputFile) {
		SketchFileIO.sketchToXML(outputFile, this);
	}

	/**
	 * Add a new, empty entity at point X, Y 
	 *
	 * @param name The name of the new entity being added
	 * @param x X Coordinate of new entity
	 * @param y Y Coordinate of new entity
	 */
	public void addNewEntity(String name, double x, double y) {
		EntityNode newEntity;

		newEntity = new EntityNode(name, (int) x, (int) y, this);
		addEntity(newEntity);
	}

	/**
	 * Add one or more entities, or an array of entities, to the graph, dealing
	 * with all of the dependencies.
	 * 
	 * @param theEntities one or more EntityNodes (or an array of entitynodes) to be added
	 */
	public void addEntity(EntityNode... theEntities) {
		addEntity(Arrays.asList(theEntities));
	}

	/**
	 * Adds a collection (set, list, etc.) of EntityNodes to the graph.
	 */
	public void addEntity(final Collection<EntityNode> theEntities) {
		//Push loading state
		_stateManager.pushState(new LoadingState(this));

		GraphLayoutCache glc = getGraphLayoutCache();

		model.beginUpdate();

		for (EntityNode node: theEntities) {
			// Set the on-screen position of our entity to the attributes of the entity
			AttributeMap nAttribs = node.getAttributes();
			GraphConstants.setAutoSize(nAttribs, true);
			GraphConstants.setBounds(nAttribs, new Rectangle2D.Double(node.getX(), node.getY(), 0, 0));

			// Add our entity to the graph				
			glc.insert(node);

			// Add our entity to our table of entities
			_entityNodes.put(node.toString(), node);

			// Add Entity to tree
			_sketchFrame.getInfoTreeUI().addEntity(node);
		}

		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				for (EntityNode node: theEntities)
					_entityNodes.remove(node.toString());
			}
			public void redo() {
				super.redo();
				for (EntityNode node: theEntities)
					_entityNodes.put(node.toString(), node);
			}
		});

		model.endUpdate();

		// Reload the graph to respect the new changes
		glc.reload();

		//Pop state
		_stateManager.popState();
	}

	/**
	 * Adds one or more edges (or an array of edges) to the sketch.
	 * 
	 * @param inEdges one or more SketchEdges (or an array of sketchedges) to be added
	 */
	public void addEdge(SketchEdge... inEdges) {
		addEdge(Arrays.asList(inEdges));
	}

	/**
	 * Adds a collection (set, list, etc.) of edges to the sketch.
	 *
	 * @param inEdges Collection of SketchEdges to add
	 */
	public void addEdge(final Collection<SketchEdge> inEdges) {
		//Push loading state
		_stateManager.pushState(new LoadingState(this));

		GraphLayoutCache glc = getGraphLayoutCache();
		model.beginUpdate();

		for (SketchEdge edge: inEdges) {
			// Add our entity to the graph				
			glc.insert(edge);
			_edges.put(edge.getName(), edge);
		}

		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				for (SketchEdge edge: inEdges) _edges.remove(edge.getName());
			}
			public void redo() {
				super.redo();
				for (SketchEdge edge: inEdges) _edges.put(edge.getName(), edge);
			}
		});

		model.endUpdate();

		//Pop state
		_stateManager.popState();

		refresh();
		_theOverview.refresh();
	}

	/**
	 * Add a new constraint to the sketch
	 * 
	 * @param c The constraint to be added
	 */
	public void addNewConstraint(Constraint... c){
		addNewConstraint(Arrays.asList(c));
	}
	public void addNewConstraint(final Collection<Constraint> c) {
		model.beginUpdate();

		for (Constraint con: c) {
			addConstraint(con);
		}

		model.endUpdate();

		refresh();
		_theOverview.refresh();
	}
	/**
	 * Adds a constraint to the sketch. This will register the constraint in the 
	 * constraint list, as well as adding a visual representation of the constraint
	 * to the graph.
	 * 
	 * @param c The constraint to be added.
	 */
	public void addConstraint(final Constraint c) {
		//Push loading state
		_stateManager.pushState(new LoadingState(this));

		model.beginUpdate();
		_constraints.addLast(c);
		
		c.setVisible(c.isVisible());
		_sketchFrame.getInfoTreeUI().addConstraint(c);

		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_constraints.removeLast();
			}
			public void redo() {
				super.redo();
				_constraints.addLast(c);
			}
		});

		model.endUpdate();

		//Pop state
		_stateManager.popState();
	}

	/**
	* Returns the next available name, starting from 'NewEntity0', so we don't get duplicates.
	* 
	* @return The next new name.
	*/
	public String getNewName() {
		return getNewName("NewEntity0");
	}

	/**
	 * Returns the next available name, starting from the passed-in name.
	 *
	 * @param name the initial name to try; if it exists, we add/increment a number until it's unique
	 * @return the first available unused name
	 */
	public String getNewName(String name) {
		while (isNameUsed(name))
			name = EasikTools.incrementName(name);

		return name;
	}

	/**
	* Checks to see if a name is in use, so that we will not have several instances
	* at once.
	* 
	* @param inName The desired new name to check against
	* @return Is it used or not.
	*/
	public boolean isNameUsed(String inName) {
		
		Set<String> keys = _entityNodes.keySet();
		for (String name: keys){
			if(name.equalsIgnoreCase(inName))
				return true;
		}
		for (ViewNode vn : _views.values()){
			for (QueryNode qn: vn.getView().getQueryNodes()){
				if(qn.getName().equalsIgnoreCase(inName))
					return true;
			}
		}
		
		return false;
	}

	/**
	* Returns the first available unique edge identifier
	* @param isInjective true if the edge is injective, false otherwise
	* @return The next id
	*/
	public String getNewEdgeName(boolean isInjective) {
		int currentID = 0;
		boolean foundOne = false;

		String prefix = isInjective ? "isA_" : "f";

		while (!foundOne) {
			currentID++;
			if (!isEdgeNameUsed(prefix + currentID)) {
				foundOne = true;
			}
		}
		return prefix + currentID;
	}

	/**
	 * Checks to see if a name is in use, so that we will not have several instances
	 * at once. For edges.
	 * 
	 * @param inName The desired new edge name to check against
	 * @return Is it used or not.
	 */
	public boolean isEdgeNameUsed(String inName) {
		
		Set<String> edges = _edges.keySet();
		for(String e : edges){
			if(e.equalsIgnoreCase(inName))
				return true;
		}
		
		return false;
	}

	/**
	 * Determines if this sketch contains any elements.
	 *
	 * @return True if this sketch contains any elements, and false otherwise
	 */
	public boolean isEmpty(){
		return getRoots().length == 0;
	}

	/**
	 * Getter method for the overview in which this sketch exists.
	 *
	 * @return The overview in which this sketch exists
	 */
	public Overview getOverview(){
		return _theOverview;
	}

	/**
	 * Takes a snapshot of the current sketch and returns it.  We override
	 * JGraph's default getImage method to make sure the sketch is displayable,
	 * setting it up propertly if it isn't.
	 *
	 * @param bg background color of the sketch, or null for transparent
	 * @param inset the size of an empty border to add around the sketch
	 * @return an Image of the current sketch
	 */
	public BufferedImage getImage(Color bg, int inset) {
		if( ! _sketchFrame.isDisplayable()){
			// set preferred size to current size, and then pack to make sketch displayable.
			// This is required for JGraph.getImage(..) to work properly
			_sketchFrame.setPreferredSize(_sketchFrame.getSize());
			_sketchFrame.pack();
		}

		return super.getImage(bg, inset);
	}

	/**
	 * Takes a snapshot of the current sketch, sends the image to scaleImage(..), and sets in
	 * the node representing this sketch.
	 */
	public void updateThumb(){
		Image temp = (Image) getImage(Easik.getInstance().getSettings().getColor("edit_canvas_background"), 0);

		if(temp != null) {
			temp = scaleThumb((BufferedImage)temp);
		}

		_sketchFrame.getNode().setThumbnail(temp);
	}

	/**
	 * Scales image by factor specified for x and y directions.
	 *
	 * @param inThumb The image to be scaled
	 *
	 * @return The scaled image
	 */
	private static Image scaleThumb(BufferedImage inThumb){
		double scaleFactor = Double.parseDouble(Easik.getInstance().getSettings().getProperty("thumb_scale_factor", "0.5"));
		return inThumb.getScaledInstance((int)(inThumb.getWidth()*scaleFactor), (int)(inThumb.getHeight()*scaleFactor), Image.SCALE_SMOOTH);
	}

	/**
	 * Returns true if this Sketch already has a database driver, false otherwise.  If this
	 * returns false, a call to getDatabaseDriver will create a new database driver object,
	 * otherwise it will simply return the existing one.
	 */
	public boolean hasDatabaseDriver() {
		return _dbd != null;
	}

	/**
	 * Returns true if this Sketch has a database driver and it has an active connection. The same
	 * can be accomplished by testing if this.getDatabaseDriver() has as connection, but this method
	 * eliminates the need for a try/catch mess.
	 */
	public boolean hasConnection(){
		if( _dbd != null){
			return _dbd.hasConnection();
		}
		return false;
	}

	/**
	 * Gets an update monitor for this sketch. As update monitors need a database driver to perform their
	 * operations, an attempt will be made to establish a driver if one not already present.
	 * @return An update monitor that can be used to perform updates on the database represented by
	 * 		sketch. If a database driver cannot be established, returns null
	 */
	public UpdateMonitor getUpdateMonitor(){
		try{
			if( getDatabaseDriver() == null )
				return null;
		}
		catch(SQLException e){
			JOptionPane.showMessageDialog(_sketchFrame, "Error getting database driver: " + e.getMessage() );
			return null;
		}
		catch(DatabaseDriver.LoadException e){
			JOptionPane.showMessageDialog(_sketchFrame, "Error getting database driver: " + e.getMessage() );
			return null;
		}

		if( _updateMonitor == null ){
			_updateMonitor = new UpdateMonitor(this, _dbd);
		}
		return _updateMonitor;
	}

	/**
	 * Allows the user to update the current connection properties. The connection options
	 * dialog is given to the user with the current connection properties (excluding user's
	 * password) as the field defaults. If the newly defined connection fails to load, the
	 * original connection is kept.
	 * If our driver is initially null, this method accomplishes the same thing as getDatabaseDriver().
	 *
	 * @throws DatabaseDriver.LoadException if the driver cannot be loaded
	 * @throws SQLException if an SQL exception occurs while attempting to connect
	 */
	public void updateDatabaseDriver(){
		DatabaseDriver oldDriver = _dbd;

		String choice = getDBType();

		if (choice != null) {
			DatabaseOptions dbopts = new DatabaseOptions(choice, _dbd, _sketchFrame);
			if(!dbopts.isAccepted())
				return;

			try{
				DatabaseDriver newDriver = DatabaseDriver.loadDriver(choice, dbopts.getParams());
				_dbd.getConnection();
				_dbd = newDriver;
				if (setConnectionParams(dbopts.getParams()))
					setDirty();
			}
			catch(DatabaseDriver.LoadException e){
				_dbd = oldDriver;
				JOptionPane.showMessageDialog(null, "Settings left unchanged: " + e.getMessage());
			}
			catch(SQLException e){
				_dbd = oldDriver;
				JOptionPane.showMessageDialog(null, "Settings left unchanged: " + e.getMessage());
			}
			_sketchFrame.setConnectionStatus(_dbd);
		}
	}

	/**
	 * Prompts user for database type and then passes off to getDatabaseDriver(String type)
	 */
	public DatabaseDriver getDatabaseDriver() throws DatabaseDriver.LoadException, SQLException {
		if( _dbd != null )
			return _dbd;
		else
			return getDatabaseDriver(getDBType());
	}

	/**
	 * Returns the current database driver, if already set, otherwise triggers a
	 * modal dialog box for connection options and creates a new one using 
	 * @param The SQL dialect to be used if new database driver must be created
	 * @return the database driver configured with the user's parameter, or null if the user cancelled the parameter dialog
	 */
	public DatabaseDriver getDatabaseDriver(String type) throws DatabaseDriver.LoadException, SQLException {
		if (_dbd == null && type != null) {
			DatabaseOptions dbopts = new DatabaseOptions(type, _sketchFrame);
			if(dbopts.isAccepted()){
				_dbd = DatabaseDriver.loadDriver(type, dbopts.getParams());
				if (setConnectionParams(dbopts.getParams()))
					setDirty();
			}
		}
		_sketchFrame.setConnectionStatus(_dbd);
		return _dbd;
	}

	public void setDatabaseDriver(DatabaseDriver dbd) {
		_dbd = dbd;
	}

	public void cleanDatabaseDriver() {
		_dbd = null;
		_sketchFrame.setConnectionStatus(_dbd);
	}

	/** Prompts the user for a database type, returns the result (or null if they cancelled) */
	public String getDBType() {
		String[] activeTypes = DatabaseDriver.availableDrivers();
		int defaultIndex = 0;
		String defType = _connParams.get("driver");
		if (defType == null) defType = Easik.getInstance().getSettings().getProperty("sql_driver");

		if (defType != null) {
			for (int i = 0; i < activeTypes.length; i++) {
				if (defType.equals(activeTypes[i])) {
					defaultIndex = i;
					break;
				}
			}
		}

		return (String) JOptionPane.showInputDialog(getFrame(), "Please select the database type to use", "Choose Database Type",
				JOptionPane.INFORMATION_MESSAGE, null, activeTypes, activeTypes[defaultIndex]);
	}

	/**
	 * Stores the map of database connection options reflecting the last
	 * connection used for this sketch.  We don't allow "password" to be
	 * stored.
	 *
	 * @param params Map of connection parameters to store
	 * @return true if the connection parameters are changed.
	 * @see getConnectionParams()
	 */
	public boolean setConnectionParams(Map<String, String> params) {
		boolean changed = false;
		for (String k: params.keySet()) {
			if (setConnectionParam(k, params.get(k)))
				changed = true;
		}
		return changed;
	}

	/**
	 * Clears the current map of database connection options.
	 */
	public void clearConnectionParams() {
		_connParams.clear();
	}

	/**
	 * Sets a single connection parameter in the current database connection
	 * options hashmap.  Returns true if the value is new (that is, it didn't
	 * exist, or it was set to something else).
	 *
	 * @param key the connection parameter name
	 * @param value the connection parameter value
	 * @return true if the connection params are now different as a result of this change
	 * @see setConnectionParams(Map)
	 */
	public boolean setConnectionParam(String key, String value) {
		if (!key.equals("password")) {
			String old = _connParams.get(key);
			_connParams.put(key, value);
			if (old == null || !old.equals(value))
				return true;
		}
		return false;
	}

	/**
	 * Returns a map of saved database connection options reflecting the last
	 * connection used for this sketch (saved by setConnectionParams()).
	 *
	 * @return Map of String, String pairs
	 * @see setConnectionParams(Map)
	 */
	public Map<String, String> getConnectionParams() {
		return Collections.unmodifiableMap(_connParams);
	}

	/**
	 * Gets the name of this sketch. Simply a shortcut for _docInfo.getName().
	 */
	public String getName(){
		return _docInfo.getName();
	}

	/**
	 * Sets the default cascade mode for new non-partial edges of this sketch.
	 *
	 * @param cascade the new Cascade mode, e.g. <code>Cascade.RESTRICT</code>.
	 * 		<code>Cascade.SET_NULL</code> is not permitted: if passed, it will
	 * 		be ignored.
	 */
	public void setDefaultCascading(Cascade newDefault) {
		if (newDefault == Cascade.SET_NULL)
			return; // Impossible choice.

		_defaultCascade = newDefault;
	}
	/**
	 * Returns the current default cascade mode for new non-partial edges of
	 * the sketch.
	 *
	 * @return the Cascade value for new non-partial edges
	 */
	public Cascade getDefaultCascading() {
		return _defaultCascade;
	}

	/**
	 * Sets the default cascade mode for new partial edges of this sketch.
	 *
	 * @param cascade the new Cascade mode, e.g. <code>Cascade.SET_NULL</code>.
	 */
	public void setDefaultPartialCascading(Cascade newDefault) {
		_defaultPartialCascade = newDefault;
	}
	/**
	 * Returns the current default cascade mode for new non-partial edges of
	 * the sketch.
	 *
	 * @return the Cascade value for new partial edges
	 */
	public Cascade getDefaultPartialCascading() {
		return _defaultPartialCascade;
	}

	/**
	 * Used to mark a sketch as dirty.  When called, we mark the Overview as
	 * dirty, and update the last modification time of the sketch to the
	 * current time.
	 */
	public void setDirty() {
		getOverview().setDirty(true);
		getOverview().refresh(_sketchFrame.getNode());
		getDocInfo().updateModificationDate();
	}

	/**
	 * Sets our sync flag flag. This indicates whether or not we are synced
	 * with a database and wish to lock editing.
	 * @param state The state of our exported flag.
	 */
	public void setSynced(boolean state){
		if ( state == false ){
			cleanDatabaseDriver();
		}
		_syncLock = state;
	}

	/**
	 * Returns the state of our sync flag. Any time a command would be issued
	 * to edit this sketch, a call should be made to this method. If it returns
	 * true, editing cannot occur safely.
	 * @return A flag indicating if this sketch has been synced with a
	 * database. If true, any changes to the sketch may destroy the link to the
	 * database.
	 */
	public boolean isSynced(){
		return _syncLock;
	}

	/**
	 * Tries random positions to find location for a new node to be placed in
	 * an effort to avoid hidden nodes. Gives up afer a specified number of
	 * tries, accepting the random position.  
	 * @param tries The number of random tries to get a new location before
	 * giving up.  
	 * @return The point deemed acceptable for placement of a new node.
	 */
	public Point getNewPosition(int tries){
		Random r = new Random();
		int w = getWidth() - 120;
		int h = getHeight() - 40;
		Point p = new Point(r.nextInt(w), r.nextInt(h));
		for ( int i=0; i<tries; i++ ) {
			if ( getFirstCellForLocation(p.getX(), p.getY()) == null )
				return p;
			p = new Point(r.nextInt(w), r.nextInt(h));
		}
		return p;
	}

	/**
	 * Returns the SketchGraphModel for this Sketch.  This is a simple wrapper
	 * + cast around the parent's getModel() method.
	 */
	public SketchGraphModel getSketchModel() {
		return (SketchGraphModel) getModel();
	}
}
