package easik.sketch;

import java.util.LinkedList;

import easik.sketch.states.BasicEditingState;
import easik.sketch.states.SketchState;

/**
 *	The state manager is a class that keeps track of the current state of the 
 *	program. Different states are used in order to interpret more complicated
 *	inputs. An example of this would be the process needed to create a 
 *	commutative diagram. First there are several clicks required to select each
 *	path. Paths can listen to what's going on with the UI, and can change and 
 *	collect information that way.
 * 
 * The state manager works like a stack. New states are pushed on top and then
 * the state manager forwards any important messages up to the active state.
 *  
 * @author Rob Fletcher 2005
 * @author Vera Ranieri 2006
 * @version 2006-05-30 Vera Ranieri
 */
public class SketchStateManager {
	/**
	 * Boolean to record whether the user has pressed the <it>Finish</it> button.
	 */
	private boolean _finished;
	/**
	 * The stack.
	 */
	private LinkedList<SketchState> _states = new LinkedList<SketchState>();
	/**
	 * The Sketch this state manager is monitoring
	 */
	private Sketch _ourSketch;
	
	/**
	 * Initialize the state manager
	 * @param inApp The current instance of this application
	 */
	public SketchStateManager(Sketch inSketch) {
		_ourSketch = inSketch;
		_finished = false;
	}
	/**
	 *
	 */
	public void initialize(){
		pushState( new BasicEditingState(_ourSketch));
	}

	/**
	 * Hook for informing the current active state that the selection has 
	 * been updated
	 */
	public void selectionUpdated() {		
		peekState().selectionUpdated();		
	}

	/**
	 * Hook for getting the currently-selectable graph elements.
	 */
	public Object[] getSelectables() {
		return peekState().getSelectables();
	}

	/**
	 * Non destructively take a look at the top item on the stack.
	 * @return The top state
	 */
	public SketchState peekState() {
		return _states.getFirst();
	}

	/**
	 * Pop the state off of the top of the stack. After the state has been popped
	 * the 'popped' hook is called. 
	 * @return The popped state
	 */
	public SketchState popState() {
		SketchState toReturn = (SketchState)_states.removeFirst() ;		
		// Inform the state it has been popped
		toReturn.poppedOff();
		// Inform the new top state that it is in the spotlight
		peekState().gotFocus();		
		return toReturn;
	}
	
	/**
	 * Pushes a state onto the stack, calls the 'pushed on' hook after being
	 * added.
	 * @param toPush The new state being pushed on
	 */
	
	public void pushState( SketchState toPush ) {
		_states.addFirst( toPush );
		toPush.gotFocus();		
		toPush.pushedOn();
	}
	
	/**
	 * Informs the top element that the next button has been clicked 
	 *
	 */
	public void nextClicked() {
		peekState().nextClicked();		
	}
	
	/**
	 * Informs the top element that the cancel button has been clicked. 
	 *
	 */
	public void cancelClicked() {
		peekState().cancelClicked();		
	}
	
	/**
	 * Informs the top element that the finish button has been clicked.
	 *
	 *@since 2006-05-30 Vera Ranieri
	 */
	public void finishClicked(){
		_finished = true;
		peekState().finishClicked();
	}
		
	/**
	 * Gets the value currently stored in <it>finsihed</it>
	 * @return True if the user has selected finished, false otherwise.
	 * @since 2006-05-30 Vera Ranieri
	 */
	public boolean getFinished(){
		return _finished;
	}
	
	/**
	 * Resets the value of <it>finsihed</it> to false
	 * @since 2006-05-30 Vera Ranieri
	 */
	public void resetFinished(){
		_finished = false;
	}
}
