package easik.sketch.attribute;

import javax.swing.tree.DefaultMutableTreeNode;

import easik.database.types.*;
import easik.sketch.keys.UniqueIndexable;
import easik.database.types.Type;
import easik.sketch.vertex.EntityNode;


/**
 * This class is used to store information related to entity attributes.
 */
public class EntityAttribute extends DefaultMutableTreeNode implements UniqueIndexable {
	/**
	 * The name of the attribute
	 */
	private String _name;
	/**
	 * The data type of the attribute
	 */
	private Type _dataType;
	/**
	 * The entity node this attribute belongs to
	 */
	private EntityNode _entity;

	/**
	 * Default Constructor
	 * 
	 * @param inName Name of attribute
	 * @param inDataType Type of attribute
	 * @see easik.database.types.Type
	 */
	public EntityAttribute(String inName, Type inDataType, EntityNode inEntity) {
		super();
		_name = inName;
		_dataType = inDataType;
		_entity = inEntity;
	}
	
	/**
	 * Overwrite standard toString function.
	 * (provides functionality for inserting attributes to the tree)
	 * @return The name of this attribute formatted for display on the tree
	 */
	public String toString(){
		return _name + " {" + getType() + "}";
	}
	
	/**
	 * Sets the attribute name.
	 * 
	 * @param inName Attribute Name
	 */
	public void setName(String inName){
		_name = inName;
	}
	
	/**
	 * Returns name of attribute.
	 * 
	 * @return Name of attribute.
	 */
	public String getName(){
		return _name;
	}
	
	/**
	 * Returns data type of this attribute
	 * 
	 * @return Type object
	 * @see easik.database.types.Type
	 */
	public Type getType(){
		return _dataType;
	}
	
	/**
	 * Sets the data type of the attribute
	 * 
	 * @param inType The data type of the attribute
	 */
	public void setType(Type inType){
		_dataType = inType;
	}

	/**
	 * Returns the entity this attribute is attached to
	 */
	public EntityNode getEntity() {
		return _entity;
	}
	
	/**
	 * Check to see if an attribute name is already used in the entity
	 * 
	 * @param inNode The entity being checked against
	 * @param attName The name being checked for duplication
	 * @return true if the name is in use, false otherwise
	 */
	public static boolean isAttNameUsed(EntityNode inNode, String attName){
		//Loop through all attributes of the entity
		for (EntityAttribute curAtt: inNode.getEntityAttributes()) {
			if (curAtt.getName().toUpperCase().equals(attName.toUpperCase()))
				return true;
		}

		//No duplicate found
		return false;
	}
}
