package easik.sketch.constraint;

import java.util.ArrayList;
import java.util.List;

import easik.sketch.path.SketchPath;
import easik.sketch.Sketch;


/** 
 *	The commutative diagram is a constraint made up of a list of lists of
 *	sketch edges. These lists share a domain and codomain.
 */
public class CommutativeDiagram extends Constraint {
	/**
	 * Creates a commutative diagram from a List of paths.  Also sets details
	 * of the visual representation.
	 * 
	 * @param inPaths List of paths to be used for the commutative diagram
	 * @param x x coordinate
	 * @param y y coordinate
	 * @param isVisible If true the constraint will be shown in the graph, otherwise it will not be visible.
	 * @param inSketch the sketch to which this CD belongs
	 */
	public CommutativeDiagram(List<SketchPath> inPaths, int x, int y, boolean isVisible, Sketch inSketch) {
		super("CD", x, y, isVisible, inSketch);
		_paths = new ArrayList<SketchPath>(inPaths);
		
		addEdges();
		
		assignToEntities();
	}

	/**
	 * Creates a commutative diagram.
	 * 
	 * @param inPaths List of paths to be used for the commutative diagram
	 * @param inSketch the sketch to which this CD belongs
	 */
	public CommutativeDiagram(List<SketchPath> inPaths, Sketch inSketch) {
		super(inSketch);

		_paths = new ArrayList<SketchPath>(inPaths);

		addEdges();

		_isVisible = true;
		setName("CD");
		assignToEntities();
	}

	/**
	 * Sets the path array, updates edge list, and updates display.  This
	 * method should not be called without first checking whether the new paths
	 * are valid, using CommutativeDiagram.isCommutativeDiagram(paths)!
	 * 
	 * @param inPaths The new array of paths
	 */
	public void setPaths(List<SketchPath> inPaths){
		_paths = new ArrayList<SketchPath>(inPaths);
		 
		addEdges();
		setVisible(!isVisible());
		setVisible(!isVisible());
	}

	/**
	 * Checks whether the current selected paths form a commutative diagram.
	 * This is determined by checking whether the paths participating in the commutative
	 * diagram have the same domain and codomain, and also by making sure that each
	 * path is different.
	 *
	 * @param inPaths The list of paths participating in the commutative diagram
	 * @return true if it forms a valid commutative diagram, false otherwise
	 */
	public static boolean isCommutativeDiagram(List<SketchPath> inPaths) {
		if (inPaths.size() < 2)
			return false; // Not a CD if it's a single path
		ArrayList<SketchPath> paths = new ArrayList<SketchPath>(inPaths);
		SketchPath firstPath = paths.get(0);
		for (int i = 0; i < paths.size(); i++) {
			SketchPath path = paths.get(i);
			// Check that the domain and codomain are equal to the first path (and therefore
			// all equal to each other).
			if (i > 0) // Don't both checking the first one, obviously it is equal to itself
				if (!firstPath.getDomain().equals(path.getDomain()) || !firstPath.getCoDomain().equals(path.getCoDomain()))
					return false;

			// Now look at every following path, and make sure none is the same as this path:
			for (int j = i+1; j < paths.size(); j++) {
				SketchPath otherPath = paths.get(j);

				// We can skip the check if that path sizes aren't the same:
				if (path.getEdges().size() == otherPath.getEdges().size()) {
					boolean different = false;
					int size = path.getEdges().size();
					for (int k = 0; k < size; k++) {
						if (!path.getEdges().get(k).equals(otherPath.getEdges().get(k))) {
							// We found a difference, so we can abort this loop
							different = true;
							break;
						}
					}
					if (!different)
						return false; // Identical paths; not a valid CD.
				}
			}
		}

		// We got through all the checks, so this must be a valid CD.
		return true;
	}
}
