package easik.sketch.constraint;

import java.util.ArrayList;
import java.util.List;

import easik.sketch.path.SketchPath;
import easik.sketch.vertex.EntityNode;
import easik.sketch.Sketch;

/** 
 * Represents the product constraint. Contains a list of paths 
 * which all share a source.
 * 
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @version 2006-05-25 Vera Ranieri
 */
public class ProductConstraint extends Constraint {
	/**
	 * Product constraint constructor. Accepts a List 
	 * that can have any number of paths.  Gives default coordinates.
	 *  
	 * @param inPaths A List of SketchPaths
	 */
	public ProductConstraint(List<SketchPath> inPaths, Sketch inSketch) {

		super(inSketch);

		setName("x");
		_isVisible = true;
		_paths = new ArrayList<SketchPath>(inPaths);
		
		addEdges();
		assignToEntities();
	}

	/**
	 * Product constraint constructor.  Accepts coordinates in constructor
	 * to allow for user-defined visual placement.
	 * 
	 * @param inPaths A List of SketchPaths
	 * @param x X coordinate of visual on graph
	 * @param y Y coordinate of visual on graph
	 * @param isVisible If the constraint is visible in the graph or not
	 */
	public ProductConstraint(List<SketchPath> inPaths, int x, int y, boolean isVisible, Sketch inSketch) {
		super("x",x,y,isVisible,inSketch);
		_paths = new ArrayList<SketchPath>(inPaths);
		
		addEdges();
		assignToEntities();
	}
	
	/**
	 * Sets the path array, updates edge list, and updates display
	 * 
	 * @param inPaths The new array of paths
	 */
	public void setPaths(List<SketchPath> inPaths){
		_paths = new ArrayList<SketchPath>(inPaths);
		 
		addEdges();
		setVisible(!isVisible());
		setVisible(!isVisible());
	}
	
	
	/**
	 * Method to determine whether a given set of paths could legally create a product constraint.
	 * 
	 * @param paths The paths potentially involved in a product constraint
	 * @return True if a legal path configuration, false otherwise
	 * @since 2006-05-25 Vera Ranieri
	 */
	public static boolean isProductConstraint(List<SketchPath> paths){
		
		//We must have a bunch of injective arrows aimed at the same spot
		EntityNode sharedSource = null;
		int pathsInvolved = 0;

		// Search through the selection...
		for (SketchPath currentPath: paths) {
				// First source we come across becomes our 'sharedSource'
				if (sharedSource == null)
					sharedSource = (EntityNode) currentPath.getDomain();
				else if (sharedSource != currentPath.getDomain())
					// If we encounter a source which is not the sharedSource, fail			
					return false;
				if (sharedSource.equals(currentPath.getCoDomain()))
					// In a loop
					return false;

				// Otherwise, we're fine, add the edge
				pathsInvolved++;
		}

		return pathsInvolved >= 2;
	}
}
