package easik.sketch.constraint;

import java.util.ArrayList;
import java.util.List;
import java.util.LinkedList;

import easik.sketch.Sketch;
import easik.sketch.path.SketchPath;
import easik.sketch.edge.SketchEdge;
import easik.sketch.vertex.EntityNode;


/** 
 * Class representing a pullback constraint.  This requires 4 paths: two from
 * the pullback node to (separate!) intermediate nodes, and two from the
 * intermediate nodes to the (same) target node.
 *
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @author Jason Rhinelander 2008
 */
/* To aid in visualizing the procedure, we describe paths as top/bottom/left/right; 
 * those are obviously arbitrary designations, but are based on this sort of visual
 * representation:
 *     P
 *    / \
 *   A   B
 *    \ /
 *     C
 *
 * where "P" is the pullback, A and B are intermediate nodes, and C is the target node.
 */
public class PullbackConstraint extends Constraint {

	/**
	 * Takes a List of paths, makes them a pullback.  The paths can be specified in
	 * any order, as long as they can be arranged into a pullback.  If the paths
	 * are invalid (that is, they cannot form a pullback), a
	 * ConstraintException is thrown.
	 * 
	 * @param inPaths A List of SketchPaths 
	 * @param inSketch the Sketch this constraint belongs to
	 * @throws ConstraintException if the paths do not form a valid pullback constraint
	 */
	public PullbackConstraint(List<SketchPath> inPaths, Sketch inSketch) throws ConstraintException {
		super(inSketch);
		setName("PB");
		_isVisible = true;
		_paths = new ArrayList<SketchPath>(inPaths);
		if (!isPullbackConstraint(_paths)) // This method rearranges _paths
			throw new ConstraintException("Unable to create pullback constraint: specified paths do not form a valid pullback constraint");

		addEdges();
		assignToEntities();
	}

	/**
	 * Takes an ArrayList of paths, makes them a pullback.
	 * 
	 * @param paths An ArrayList of SketchPaths 
	 * @param x X coordinate of visual aid
	 * @param y Y coordinate of visual aid
	 * @param isVisible If the constraint is visible in the graph or not
	 */
	public PullbackConstraint(ArrayList<SketchPath> paths, int x, int y, boolean isVisible, Sketch inSketch) throws ConstraintException {
		super("PB", x, y, isVisible, inSketch);

		_paths = new ArrayList<SketchPath>(paths);
		if (!isPullbackConstraint(_paths)) // This method rearranges _paths
			throw new ConstraintException("Unable to create pullback constraint: specified paths do not form a valid pullback constraint");

		addEdges();
		assignToEntities();
	}

	/**
	 * Returns the "source" node of the pullback--that is, the pullback node itself.
	 *
	 * @return EntityNode the pullback node
	 */
	public EntityNode getSource() {
		return _paths.get(0).getDomain();
	}

	/**
	 * Returns the "target" node of the pullback (that is, the codomain of both
	 * pullback paths).
	 *
	 * @return EntityNode the pullback target node
	 */
	public EntityNode getTarget() {
		return _paths.get(_paths.size()-1).getCoDomain();
	}

	/**
	 * Returns the projection path from the pullback to one of the source nodes.  The codomain
	 * of this path will be the domain of the path returned by getTargetPathA().
	 */
	public SketchPath getProjectionA() {
		return _paths.get(0);
	}

	/**
	 * Returns the target path from one of the source nodes to the target node.  The domain
	 * of this path will be the codomain of the path returned by getProjectionA().
	 */
	public SketchPath getTargetPathA() {
		return _paths.get(1);
	}

	/**
	 * Returns the projection path from the pullback to one of the source nodes.  The codomain
	 * of this path will be the domain of the path returned by getTargetPathB().
	 */
	public SketchPath getProjectionB() {
		return _paths.get(2);
	}

	/**
	 * Returns the target path from one of the source nodes to the target node.  The domain
	 * of this path will be the codomain of the path returned by getProjectionB().
	 */
	public SketchPath getTargetPathB() {
		return _paths.get(3);
	}

	/**
	 * Returns one of the <b>full</b> paths from pullback to target.  This is
	 * the combination of *one* of the pair of paths from the pullback node to
	 * target node.  The returned path is the path consisting of paths 0 and 1.
	 *
	 * The returned path is equivalent to combining the paths returned by getProjectionA()
	 * and getTargetPathA().  This method is deprecated in favour of using those two calls
	 * individually.
	 *
	 * @return SketchPath created from the combination of paths from pullback to target nodes
	 */
	public SketchPath getFullPathA() {
		LinkedList<SketchEdge> edges = new LinkedList<SketchEdge>();
		edges.addAll(getProjectionA().getEdges());
		edges.addAll(getTargetPathA().getEdges());
		return new SketchPath(edges);
	}

	/**
	 * Returns one of the <b>full</b> paths from pullback to target.  This is
	 * the combination of *one* of the pair of paths from the pullback node to
	 * target node.  The returned path corresponds to the path consisting of
	 * paths 2 and 3.
	 *
	 * The returned path is equivalent to combining the paths returned by getProjectionB()
	 * and getTargetPathB().  This method is deprecated in favour of using those two calls
	 * individually.
	 *
	 * @return SketchPath created from the combination of paths from pullback to target nodes
	 */
	public SketchPath getFullPathB() {
		LinkedList<SketchEdge> edges = new LinkedList<SketchEdge>();
		edges.addAll(getProjectionB().getEdges());
		edges.addAll(getTargetPathB().getEdges());
		return new SketchPath(edges);
	}

	/**
	 * Method to determine whether a set of paths potentially forming a pullback constraint could legally
	 * form a pullback constraint.  Additionally, if the pullback constraint is valid, this rearranges the
	 * passed-in list to be in the appropriate order: path elements 0 and 1 form one path, elements 2 and
	 * 3 form the other path.  In other words, the domains of elements 0 and 2 are the same, and the
	 * codomains of elements 1 and 3 are the same.
	 * 
	 * Note: This method does not currently care about the order that the paths are selected.
	 * 
	 * @param paths The set of potential paths
	 * @return True if a legal pullback configuration, false otherwise.
	 * @since 2006-05-25 Vera Ranieri
	 */
	public static boolean isPullbackConstraint(List<SketchPath> paths){

		if (paths.size() !=4)
			return false;

		for (SketchPath p: paths)
			if (p == null) return false;

		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				if (j == i) continue;

				for (int k = 0; k < 4; k++) {
					if (k == i || k == j) continue;

					for (int l = 0; l < 4; l++) {
						if (l == k || l == j || l == i) continue;

						SketchPath pathA = paths.get(i), pathB = paths.get(j), pathC = paths.get(k), pathD = paths.get(l);
						EntityNode domA = pathA.getDomain(), domB = pathB.getDomain(), domC = pathC.getDomain(), domD = pathD.getDomain();
						EntityNode codoA = pathA.getCoDomain(), codoB = pathB.getCoDomain(), codoC = pathC.getCoDomain(), codoD = pathD.getCoDomain();
						try{
							if (
								codoA == domB && // "Left" paths connect
								codoC == domD && // "Right" paths connect
								codoB == codoD && // "Bottom" paths point to same thing
								domA == domC // "Top" paths come from the same thing
							   ) {
								/*
								   []
							      /  \
							    B/    \D
							    /      \
							   []      []
							    \      /
						        A\    /C
							      \  /
							       []
								   Make sure that injectiveness makes sense
								*/
								if(! ( !pathB.getEdges().getFirst().isInjective()
											||
											pathC.getEdges().getFirst().isInjective()
											&&
											!pathD.getEdges().getFirst().isInjective()
											||
											pathA.getEdges().getFirst().isInjective() ) )
								{ return false; }

								paths.clear();
								paths.add(pathA);
								paths.add(pathB);
								paths.add(pathC);
								paths.add(pathD);
								return true;
							}
						}
						catch (Exception e){
							//TODO: better way to determine whether there is a problem with the paths (i.e. 
							// there is a path of length 0)
							return false;
						}
					}
				}
			}
		}

		return false;
	}
}
