package easik.sketch.keys;

import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Collections;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.undo.AbstractUndoableEdit;

import easik.sketch.vertex.EntityNode;


/**
 * Class used to keep track of attributes used in unique key constraints.
 * 
 * @author Kevin Green 2006
 * @since 2006-06-06 Kevin Green
 * @version 2006-07-13 Kevin Green
 */
public class UniqueKey extends DefaultMutableTreeNode {

	/**
	 * The attributes contributing to this unique key
	 */
	private LinkedHashSet<UniqueIndexable> _elements;
	/**
	 * The name of this key
	 */
	private String _keyName;
	/**
	 * The entity node to which this key applies
	 */
	private EntityNode _parent;
	
	/**
	 * Default constructor
	 * @param parent The EntityNode to which this unique key applies
	 * @param elements A List of the indexable objects (implementing UniqueIndexable) used in the key
	 * @param name The name of the key
	 */
	public UniqueKey(EntityNode parent, Collection<UniqueIndexable> elements, String name) {
		super();
		_parent = parent;
		_elements = new LinkedHashSet<UniqueIndexable>(elements);
		_keyName = name;
	}

	/**
	 * Overwriten method returns attribute list separated by semicolons
	 * 
	 * @return The name of this unique key formatted for use on the information tree
	 */
	public String toString(){
		String id = getId();
		id = _keyName + " (" + id + ")";
		return id;
	}
	
	/**
	 * Gets the id of this unique key.  The id is formatted by getting the name of each attribute and placing
	 * a comma between each name
	 * @return The id of this attribute
	 */
	public String getId() {
		StringBuilder id = new StringBuilder();
		for (UniqueIndexable elem: _elements) {
			id.append(elem.getName() + ",");
		}
		return id.substring(0, id.length()-1);
	}

	/**
	 * Returns the set of unique-indexable elements (attributes and edges) that
	 * make up this key.  The set is backed by a LinkedHashSet, and so is ordered.
	 *
	 * @return a Set of the UniqueIndexable objects (attributes/edges) used in the key
	 */
	public Set<UniqueIndexable> getElements() {
		return Collections.unmodifiableSet(_elements);
	}

	/**
	 * Sets the elements involved in this unique key to the unique-indexable
	 * attributes/edges contained in the passed-in collection (set, list,
	 * etc.).  Note that this collection will be added to a set, so duplicate
	 * EntityAttribute values will be ignored.  Also note that the order of the
	 * attributes will be preserved.
	 *
	 * @param elems Collection of UniqueIndexable-implementing objects
	 */
	public void setElements(Collection<UniqueIndexable> elems) {
		final LinkedHashSet<UniqueIndexable> oldElements = _elements;
		_elements = new LinkedHashSet<UniqueIndexable>(elems);
		final LinkedHashSet<UniqueIndexable> newElements = _elements;
		_parent.getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
			public void undo() { super.undo(); _elements = oldElements; }
			public void redo() { super.redo(); _elements = newElements; }
		});
	}

	/**
	 * Returns a UniqueIndexable element from the element list, if it contains it.  Note that this
	 * might make the unique key invalid (0 elements) or a duplicate of another key: you must take
	 * care to call UniqueKey.cleanup(node) after finished removing all elements from the node
	 * to fix up the element list.
	 * 
	 * @param elem the UniqueIndexable-implementing object to remove
	 * @return true if the element existed and was removed, false if the element did not exist
	 */
	public boolean removeElement(final UniqueIndexable elem) {
		boolean ret = _elements.remove(elem);
		if (ret) {
			_parent.getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
				public void undo() { super.undo(); _elements.add(elem); }
				public void redo() { super.redo(); _elements.remove(elem); }
			});
		}
		return ret;
	}

	/**
	 * Returns the name of the key
	 * 
	 * @return The name of the key
	 */
	public String getKeyName(){
		return _keyName;
	}
	
	/**
	 * Sets the key name to the name described in <it>inName</it>
	 * @param inName The name of this attribute
	 */
	public void setKeyName(final String inName) {
		final String oldName = _keyName;
		_keyName = inName;
		_parent.getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
			public void undo() { super.undo(); _keyName = oldName; }
			public void redo() { super.redo(); _keyName = inName; }
		});
	}

	/**
	 * Returns the parent entity for the key
	 * 
	 * @return The parent entity for the key
	 */
	public EntityNode getEntity() {
		return _parent;
	}

	/**
	 * Returns if the name of a key is already used
	 * 
	 * @param inNode The entity being checked against
	 * @param keyName The name of the key
	 * @return true if the key name is already used, false otherwise
	 */
	public static boolean isKeyNameUsed(EntityNode inNode, String keyName) {
		for (UniqueKey q: inNode.getUniqueKeys()) {
			if (q.getKeyName().toUpperCase().equals(keyName.toUpperCase()))
				return true;
		}
		return false;
	}

	/**
	 * Finds any empty or duplicate unique keys of the passed-in EntityNode and
	 * returns a list of them.  This is used by cleanup(EntityNode) to remove
	 * problem keys after removing an attribute or edge.
	 *
	 * @param inNode the node to check
	 * @return List of UniqueKey objects that should be removed to clean up the table
	 */
	public static List<UniqueKey> findBadKeys(EntityNode inNode) {
		HashSet<Set<UniqueIndexable>> seen = new HashSet<Set<UniqueIndexable>>();
		LinkedList<UniqueKey> badKeys = new LinkedList<UniqueKey>();

		for (UniqueKey q: inNode.getUniqueKeys()) {
			Set<UniqueIndexable> qElems = q.getElements();
			if (qElems.isEmpty()) // The attribute set is now empty
				badKeys.add(q);
			else if (seen.contains(qElems)) // We've already encountered this set of attributes
				badKeys.add(q);
			else // Otherwise it's good, but remember that we've seen it already
				seen.add(qElems);
		}

		return badKeys;
	}

	/**
	 * Tests to see if there are any empty or duplicate unique keys in the
	 * passed-in EntityNode, and if so, removes them.  This is typically called
	 * when removing an attribute to fix up the EntityNode, since the attribute
	 * will have also been removed from the existing unique keys.
	 *
	 * @param inNode The node whose unique keys should be cleaned
	 * @return true if any unique keys were removed
	 */
	public static boolean cleanup(EntityNode inNode) {

		List<UniqueKey> badKeys = findBadKeys(inNode);

		if (badKeys.isEmpty()) return false;

		for (UniqueKey remove: badKeys) {
			inNode.removeUniqueKey(remove);
		}

		return true;
	}
}
