package easik.sketch.states;

import java.util.ArrayList;

import javax.swing.JOptionPane;

import easik.sketch.Sketch;
import easik.sketch.constraint.Constraint;
import easik.sketch.constraint.ProductConstraint;
import easik.sketch.path.SketchPath;


/** 
 * This state makes a product constraint. When pushed, it will then push another
 * state to get a path. Once the first path is received, it will push another request for
 * a path. After adding the first path the options to either add another path or finish
 * the constraint are available.
 *
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @since 2006-05-23 Kevin Green
 * @version 2006-08-15 Kevin Green
 */
public class AddProductConstraintState extends SketchState implements PathAcceptingState {
	
	/**
	 * The paths involved in this constraint
	 */
	private ArrayList<SketchPath> _paths;
	/**
	 * Stores whether the selection of paths has finished for this constraint
	 */
	private boolean _finished;
	/**
	 * If set, we delete this constraint immediately before adding the new one.  Used when adding
	 * paths to an existing constraint.
	 */
	private ProductConstraint _replace;
	
	/**
	 * Constructor for creating a new state to collect paths and make a diagram.
	 * 
	 * @param inSketch The sketch being worked with.
	 */
	public AddProductConstraintState(Sketch inSketch) {
		super(inSketch);
		_finished = false;
		_paths = new ArrayList<SketchPath>();
	}

	/**
	 * Constructor for modifying an existing constraint.
	 *
	 * @param inSketch The sketch being worked with
	 * @param existing The existing constraint to have paths added to
	 */
	public AddProductConstraintState(Sketch inSketch, ProductConstraint existing) {
		super(inSketch);
		_finished = false;
		_paths = new ArrayList<SketchPath>();
		for (SketchPath path: existing.getPaths()) {
			_paths.add(new SketchPath(path.getEdges()));
		}
		_replace = existing;
	}

	/**
	 * Called when a new path has been completed by a state above this on the
	 * stack. If it is null, then cancel, otherwise check to see if
	 * the user selected the 'next' or 'finish' option.
	 * 
	 * @param inPath The last path in, null if it was a cancellation
	 */
	public void passPath(SketchPath inPath) {
		_finished = _ourSketch.getStateManager().getFinished();

		//Check if cancelled
		if (inPath == null) {
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			_ourSketch.getStateManager().popState();
			return;
		}
		// if we have a duplicate path, let user try again
		else if(_paths.contains(inPath)){
			JOptionPane.showMessageDialog(
					_ourSketch.getParent(),
					"Cannot select duplicate path. Try Again!",
					"Easik",
					JOptionPane.INFORMATION_MESSAGE);

			_ourSketch.getStateManager().pushState(new GetPathState(true, _paths.size() > 0, _ourSketch, _paths.get(0).getDomain(), null));
			return;
		}
		// if we are on our 2nd+ path it's invalid, let user retry
		else if(_paths.size() > 0 && !(inPath.getDomain() == _paths.get(0).getDomain()) ){
			JOptionPane.showMessageDialog(
					_ourSketch.getParent(),
					"Must select path starting from '" + _paths.get(0).getDomain() + "'. Try again!",
					"Easik",
					JOptionPane.INFORMATION_MESSAGE);

			// let user try again
			_ourSketch.getStateManager().pushState(new GetPathState(true, _paths.size() > 0, _ourSketch, _paths.get(0).getDomain(), null));
			return;
		}
		
		// if still not finished, add path and push new state to allow user to select another path
		if(!_finished){

			// if we got this far, our new path is OK
			_paths.add(inPath);
			// once we have selected our first path, we can update instructions and leave until finished or canceled
			if(_paths.size() == 1){
				_ourSketch.getFrame().setInstructionText("Select a path with domain '" +
						inPath.getDomain().getName() +
						"', then press 'Next' for another path or 'Finish' to accept.");
			}

			_ourSketch.getStateManager().pushState(new GetPathState(true, true, _ourSketch, inPath.getDomain(), null));
		}
		// add the path and create constraint
		else{
			_paths.add(inPath);
			_ourSketch.getSketchModel().cancelInsignificantUpdate();

			_ourSketch.getSketchModel().beginUpdate();
			if(addDiagram()){
				// If we're replacing a constraint, remove the old one:
				if (_replace != null) {
					_ourSketch.removeConstraint(_replace);
				}
			}
			else{
				JOptionPane.showMessageDialog(
						_ourSketch.getFrame(),
						"FIXME if you see this - we should not have been able to make it here!!",
						"Error",
						JOptionPane.ERROR_MESSAGE);
			}
			_ourSketch.getSketchModel().endUpdate();

			_ourSketch.getStateManager().popState();
		}
	}

	/**
	 * Add the diagram to the sketch
	 * @return true if the diagram was successfully added, false otherwise
	 */
	private boolean addDiagram() {
		if(ProductConstraint.isProductConstraint(_paths)){
			Constraint newDiagram = _replace != null
				? new ProductConstraint(_paths, _replace.getX(), _replace.getY(), true, _ourSketch)
				: new ProductConstraint(_paths, _ourSketch);
			_ourSketch.addNewConstraint(newDiagram);
			_ourSketch.setDirty();
			return true;
		}
		return false;
	}


	/**
	 * When this state is pushed on, it sends a message in a popup and then pushes
	 * a path collection state.
	 */
	public void pushedOn() {
		_ourSketch.clearSelection();
		_ourSketch.getStateManager().resetFinished();
		_ourSketch.getSketchModel().beginInsignificantUpdate();

		_ourSketch.getFrame().setInstructionText(
				_paths.size() >= 1
				? "Select a path with domain '" + _paths.get(0).getDomain().getName() + "', then press 'Next' for another path or 'Finish' to accept."
				: "Select a path and then press 'Next' to add it to the constraint");
		setNextButton(true);
		setFinishButton(_paths.size() >= 1);
		setCancelButton(true);
		_ourSketch.getStateManager().pushState(
				new GetPathState(true, _paths.size() >= 1, _ourSketch,
					(_paths.size() >= 1 ? _paths.get(0).getDomain() : null), null));
	}

	/**
	 * Nothing to do when it gets popped off, as it always pops itself off when completed. 
	 */
	public void poppedOff() {
	}

	/**
	 * This state is called "New Product Constraint"
	 * 
	 * @return The string literal described above.
	 */
	public String toString() {
		return "New Product Constraint";
	}

}

