package easik.sketch.states;

import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Map;
import java.util.ArrayList;

import org.jgraph.graph.GraphConstants;

import easik.sketch.Sketch;
import easik.sketch.edge.SketchEdge;
import easik.sketch.path.SketchPath;
import easik.sketch.vertex.EntityNode;

/** 
 * The GetPathState is a state which can be used to retrieve a single path from the 
 * user. It ensures that the only edges which are selectable are those which are attached
 * to the previous edge.
 *
 * @author Rob Fletcher 2005
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @version 2006-07-13 Kevin Green
 */
public class GetPathState extends SketchState {

	/** Contains elements of the sketch that are selectable */
	protected Map _selectable;

	/** Contains elements of the sketch that are not selectable */
	protected Map _unselectable;

	/** The path created from the edges chosen */
	protected SketchPath _pathToReturn;	

	/** Stores whether the next state is to add a new path */
	protected boolean _nextState;

	/** Stores whether the next state is the finish state */
	protected boolean _finishState;

	/** Stores the desired start node */
	protected EntityNode _sourceNode;

	/** Stores the desired target node */
	protected EntityNode _targetNode;

	/**
	 * Default Constructor
	 * 
	 * @param next boolean determining whether the user should be allowed to select next on this round.
	 *   True if next could be selected, false otherwise.
	 * @param finish boolean determining whether the user should be allowed to select finish on this round.  
	 *   True if finish could be selected, false otherwise.
	 * @param inSketch the sketch in which this is occurring
	 */
	public GetPathState(boolean next, boolean finish, Sketch inSketch) {
		this(next, finish, inSketch, null, null);
	}

	/**
	 * Gets a path that must begin at the specified source node.
	 *
	 * @param next boolean determining whether the user should be allowed to select next on this round.
	 *   True if next could be selected, false otherwise.
	 * @param finish boolean determining whether the user should be allowed to select finish on this round.  
	 *   True if finish could be selected, false otherwise.
	 * @param inSketch the sketch in which this is occurring
	 * @param sourceNode EntityNode at which the path must begin; only edges with this node as source will
	 *   be initially selectable.  null means the path can start anywhere.
	 * @param targetNode EntityNode at which the path must end; until a path ending at targetNode is selected,
	 *   the user will be unable to click next/finish.  null means the path can end anywhere.
	 */
	public GetPathState(boolean next, boolean finish, Sketch inSketch, EntityNode sourceNode, EntityNode targetNode) {
		super(inSketch);
		_pathToReturn = null;
		_nextState = next;
		_finishState = finish;
		_sourceNode = sourceNode;
		_targetNode = targetNode;
	}

	/**
	 * When pushed on, the first thing done is clearing the selection
	 * and then disabling selection for all items except for edges.
	 */
	public void pushedOn() {
		setNextButton(false);
		setCancelButton(true);
		setFinishButton(false);
		
		_selectable = new Hashtable();
		_unselectable = new Hashtable();

		GraphConstants.setSelectable(_selectable, true);
		GraphConstants.setSelectable(_unselectable, false);

		// Initially, we allow all edges and entities to be selected
		_ourSketch.clearSelection();

		_ourSketch.refresh();

		// Gets pushed on by all "add constraint" states. We don't want to be able to imbed adding constraints
		_ourSketch.getFrame().enableAddConstraintItems(false);
	}

	/**
	 * Returns the edges that are to be selectable.  Only edges extending from
	 * the current path are selectable; if there is no path, then any edge
	 * qualifies.
	 */
	public Object[] getSelectables() {
		ArrayList<SketchEdge> selectable = new ArrayList<SketchEdge>();

		Object[] newSelection = _ourSketch.getSelectionCells();

		if (newSelection.length == 0) {
			if (_sourceNode == null) {
				// Can start anywhere
				for (Object cell: _ourSketch.getRoots()) {
					if (cell instanceof SketchEdge)
						selectable.add((SketchEdge) cell);
				}
			}
			else {
				// Must start at _sourceNode
				selectable.addAll(_sourceNode.getOutgoingEdges());
			}
		}
		else {
			// If edges are selected, we'll grab the last edge
			SketchEdge lastEdge = (SketchEdge) newSelection[newSelection.length - 1];

			// We need that last edge to be selectable (so that you can deselect it)
			selectable.add(lastEdge);

			EntityNode goodTarget = lastEdge.getTargetEntity();

			// All of goodTarget's outgoing edges are selectable:
			selectable.addAll(goodTarget.getOutgoingEdges());
		}

		return selectable.toArray();
	}

	/**
	 * Update the selection so that the only selectable items will be those within
	 * reach of the existing edges.
	 */
	public void selectionUpdated() {
		_ourSketch.refresh();
		Object[] newSelection = _ourSketch.getSelectionCells();

		// First check to see if the selection is empty
		if (newSelection.length == 0) {
			// We can't have a next/finish button available:
			setNextButton(false);
			setFinishButton(false);
		}
		else {
			// If we have an explicit target, you can hit next only if the selected path codomain matches
			boolean codoOK = _targetNode == null
				? true // No target node: always okay
				: ((SketchEdge) newSelection[newSelection.length-1]).getTargetEntity() == _targetNode;

			//Since we have something selected, then we can enable the 'next' button, if it should be.
			setNextButton(_nextState && codoOK);

			//If the user can finish this round, then the Finish button should be activated.
			setFinishButton(_finishState && codoOK);
		}
	}

	/**
	 * When the state gets popped, then it should tell the new top item what path it had
	 * collected before being popped.
	 * Since this is called AFTER popping, it can use peek() to get the top item.	 
	 */
	public void poppedOff() {
		((PathAcceptingState)_ourSketch.getStateManager().peekState()).passPath(_pathToReturn);

		_ourSketch.getGraphLayoutCache().reload();
		_ourSketch.clearSelection();
	}

	/**
	 * If path collection has been cancelled, then pop off, and set the 
	 * path to be null.
	 */
	public void cancelClicked() {
		_pathToReturn = null;
		resetSelection();
		_ourSketch.getStateManager().popState();
	}

	/**
	 * When next is clicked, pop off after preparing an array containing the 
	 * edges in the path. Convert to the proper graph edge of the sketch.
	 */
	public void nextClicked() {
		resetSelection();
		LinkedList<SketchEdge> path = new LinkedList<SketchEdge>();
		Object[] selCells = _ourSketch.getSelectionCells();
		if (selCells.length == 1 && selCells[0] instanceof EntityNode) {
			_pathToReturn = new SketchPath((EntityNode) selCells[0]);
		}
		else {
			for (int i = 0; i < selCells.length; i++) {
				path.add((SketchEdge) selCells[i]);
			}
			_pathToReturn =	new SketchPath(path);
		}
		_ourSketch.getStateManager().popState();
	}

	/**
	 * When finish is clicked, the stack is popped off after an array 
	 * containing the path is created.
	 */
	public void finishClicked(){
		nextClicked();
	}

	/**
	 * State string identifier.
	 * 
	 * @return String literal "Select a path"
	 */
	public String toString() {
		return "Select a path";
	}

	/**
	 * Set everything to be selectable
	 */
	private void resetSelection() {
		_ourSketch.getGraphLayoutCache().edit(_ourSketch.getRoots(), _selectable);
	}
}
