package easik.sketch.vertex;

import java.awt.geom.Rectangle2D;

import easik.sketch.Sketch;

import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultPort;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.AttributeMap;

/** 
 * The superclass of all nodes on the sketch, has a name and a coordinate.
 */
public class SketchVertex extends DefaultGraphCell {
	/** The x and y coordinates we wish this vertex to be initially rendered at */
	protected int _initX, _initY;

	/** The name of this node */
	protected String _name;

	/** The port of this node that edges connect to */
	protected DefaultPort _port = new DefaultPort();

	/** The sketch to which this node belongs */
	protected Sketch _theSketch;

	/**
	 * Create a blank sketch vertex, no name, and located at (0, 0)
	 *
	 * @param inSketch the Sketch the vertex is being added to
	 */
	public SketchVertex(Sketch inSketch) {
		this("", 0, 0, inSketch);
	}
	
	/**
	 * Create a vertex with basic attributes
	 * @param name Label
	 * @param x X coordinate
	 * @param y Y coordinate
	 * @param inSketch the sketch the vertex is being added to
	 */
	public SketchVertex(String name, int x, int y, Sketch inSketch) {
		_initX = x;
		_initY = y;
		_name = name;
		_theSketch = inSketch;

		add(_port);
	}
	
	/**
	 * Mutator for the name field
	 * @param name The new name
	 */
	public void setName(String name) {
		_name = name;

		_theSketch.refresh(this);
	}
	
	/**
	 * To string method returns the name
	 * @return Current name
	 */
	public String toString() {
		return getName();
	}
	
	/**
	 * Accessor for the name field
	 * @return the name
	 */
	public String getName() {
		return _name;
	}

	/**
	 * Accessor for the X coordinate
	 * @return The x coordinate
	 */
	public int getX() {
		Rectangle2D bounds = GraphConstants.getBounds(getAttributes());
		if ( bounds != null ) 
			return (int)bounds.getX();
		else
			return _initX;
	}
	
	/**
	 * Accessor for the Y Coordinate 
	 * @return	The Y Coordinate
	 */
	public int getY() {
		Rectangle2D bounds = GraphConstants.getBounds(getAttributes());
		if (bounds != null )
			return (int)bounds.getY();
		else
			return _initY;
	}	

	/**
	 * Returns the Sketch this node belongs to.
	 *
	 * @return Sketch
	 */
	public Sketch getSketch() {
		return _theSketch;
	}

	/**
	 * Returns the DefaultPort object that edges pointing to this vertex should
	 * be attached to.
	 */
	public DefaultPort getPort() {
		return _port;
	}

	/**
	 * Overrides superclass getAttributes to override selectable status.
	 */
	public AttributeMap getAttributes() {
		AttributeMap attrs = super.getAttributes();

		// Always selectable:
		GraphConstants.setSelectable(attrs, true);

		return attrs;
	}
}
