package easik.ui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.InputEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.ImageProducer;
import java.net.URL;
import java.util.Arrays;
import java.io.File;

import javax.swing.*;
import javax.swing.event.*;

import easik.Easik;
import easik.overview.Overview;
import easik.overview.vertex.*;
import easik.sketch.Sketch;
import easik.sketch.constraint.Constraint;
import easik.sketch.vertex.*;
import easik.ui.menu.*;
import easik.ui.menu.popup.*;
import easik.ui.tree.OverviewInfoTreeUI;
import easik.ui.tree.popup.*;
import easik.os.EasikMenuAdjuster;


/** 
 * This is the swing starting point for the program. This frame contains the 
 * entire application. It is accessible through the EASIK singleton by all other
 * classes.
 * 
 * @author Rob Fletcher 2005
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @version 2006-08-04 Kevin Green
 */
public class ApplicationFrame extends EasikFrame {
	/** The popup menu for right-clicking on the overview canvas */
	private JPopupMenu _canvasPopupMenu;

	/** The popup menu for right-clicking on a sketch */
	private JPopupMenu _sketchPopupMenu;

	/** The popup menu for right-clicking on a view */
	private JPopupMenu _viewPopupMenu;

	/** The popup menu for mixed selections */
	private JPopupMenu _mixedPopupMenu;

	/** The point where the popup menu should appear */
	private Point _popupPosition;

	/** The right and main panels of the frame */
	private JPanel _rightPane;
	private JPanel _mainPane;

	/** The scroll pane containing the overview */
	private JScrollPane _overviewScrollPane;

	/** The information tree */
	private OverviewInfoTreeUI _infoTreeUI;

	/** The split pane */
	private JSplitPane _mainSplitPane;

	/** The recent files menu created from the recent files stored in the ini file */
	private JMenu _recentFilesMenu;

	/** Various menu items */
	private JMenuItem _AddSketchPopItem, _ImportSketchPopItem, _NewViewPopItem, _NewViewMenuItem,
			_OpenSketchPopItem, _OpenSketchPopDefItem, _OpenSketchManipulatePopItem, _OpenSketchManipulatePopDefItem,
			_AddInjEdgePopItem, _RenamePopItem, _DeleteSketchPopItem,
			_ExportSketchPopItem, _ExportSketchSQLPopItem, _DeleteMixedPopItem,
			_AddAttributePopItem, _AddUniqueKeyPopItem, _HideConstraintPopItem,
			_NextPopItem, _FinishPopItem, _CancelPopItem,
			_AddPathPopItem, _ProgramSettingsItem, _OpenViewPopItem, _RenameViewPopItem, _DeleteViewPopItem;

	/** The Overview under this frame */
	private Overview _overview;

	/** The EASIK logo image (reused by SketchFrames and the about screen) */
	public static Image EASIK_Icon = null;

	/** Stores the object we call when creating a new window, to adjust its menus in an OS-specific way */
	private EasikMenuAdjuster _menuAdj;

	/**
	 *	Creates the application frame, sets some sizes, sets up the properties
	 *	builds the menus, and lays out the swing components.
	 */
	public ApplicationFrame() {
		super("EASIK - Untitled");

		int defaultWidth = Integer.parseInt(_settings.getProperty("overview_display_width", "700"));
		int defaultHeight = Integer.parseInt(_settings.getProperty("overview_display_height", "500"));
		int defaultLocationX = Integer.parseInt(_settings.getProperty("overview_frame_location_x", "0"));
		int defaultLocationY = Integer.parseInt(_settings.getProperty("overview_frame_location_y", "25"));
		this.setLocation(defaultLocationX, defaultLocationY);
		this.setSize(defaultWidth, defaultHeight);
		
		try{
			URL easikLogo = Easik.class.getResource("easik-icon.png");
			EASIK_Icon = Toolkit.getDefaultToolkit().createImage((ImageProducer) easikLogo.getContent());
			this.setIconImage(EASIK_Icon);
		}
		catch(Exception e){
			System.out.println("Error loading image");
		}
		
		//Initialize all variables
		_overview = new Overview(this);
		_canvasPopupMenu = new JPopupMenu();
		_sketchPopupMenu = new JPopupMenu();
		_viewPopupMenu = new JPopupMenu();
		_mixedPopupMenu = new JPopupMenu();
		_popupPosition = new Point(0, 0);
		_rightPane = new JPanel();
		_mainPane = new JPanel();
		_infoTreeUI = new OverviewInfoTreeUI(this);
		
		_infoTreeUI.refreshTree();
		
		//Build Menus
		buildMenu();
		buildPopupMenu();
		
		//Setup Sketch Pane
		_overviewScrollPane = new JScrollPane(_overview);
		_overviewScrollPane.setMinimumSize(new Dimension(300, 300));

		//Setup entire right pane
		_rightPane.setLayout(new BorderLayout());
		_treeName = new JLabel(_overview.getDocInfo().getName());
		_rightPane.add(_treeName, BorderLayout.NORTH);
		_rightPane.add(_infoTreeUI, BorderLayout.CENTER);
		
		//Setup main pane
		_mainPane.setLayout(new BorderLayout());
		_mainPane.add(_overviewScrollPane, BorderLayout.CENTER);
		
		//Setup enitre window
		_mainSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _mainPane, _rightPane);
		int dividerPos = Integer.parseInt(_settings.getProperty("overview_divider_position", "" + (defaultWidth - 255)));
		_mainSplitPane.setDividerLocation(dividerPos);
		_mainSplitPane.setDividerSize(10);
		_mainSplitPane.setResizeWeight(1);
		_mainSplitPane.setOneTouchExpandable(true);
		_mainSplitPane.setContinuousLayout(true);
		getContentPane().add(_mainSplitPane, BorderLayout.CENTER);
	}

	/**
	 * Trys to close window as long it is not dirty.
	 */
	public void closeWindow(){
		boolean proceed = true;
		if (_overview.getDirty()) {

			String[] choices = new String[] {
				"Save and Exit",
				"Exit without Saving",
				"Cancel"
			};
			String title = _overview.getDocInfo().getName();
			if (title == null || title.equals("")) title = "Untitled";
			int result = JOptionPane.showOptionDialog(
					this,
					"The document \"" + title + "\" has unsaved changes.",
					"Warning",
					JOptionPane.YES_NO_CANCEL_OPTION,
					JOptionPane.WARNING_MESSAGE,
					null,
					choices,
					choices[0]);
			if (result == 0) // Save and exit
				proceed = FileSaveAction.saveFile(this); // If this returns false, the save failed.
			else if (result == 1) // Exit without saving
				proceed = true;
			else // Cancel, or closed the dialog
				proceed = false;
		}
		if (proceed) {
			// We're exiting, so update the width/height of the main window, and save the settings
			_settings.setProperty("overview_display_width", "" + getWidth());
			_settings.setProperty("overview_display_height", "" + getHeight());
			_settings.setProperty("overview_divider_position", "" + _mainSplitPane.getDividerLocation());
			_settings.setProperty("overview_frame_location_x", "" + getX());
			_settings.setProperty("overview_frame_location_y", "" + getY());

			_settings.store();

			System.exit(0);
		}
	}

	/**
	 * Returns the Overview
	 * 
	 * @return The overview
	 */
	public Overview getOverview() {
		return _overview;
	}
	
	/**
	 * Returns the InfoTreeUI object
	 * 
	 * @return The InfoTreeUI object
	 */
	public OverviewInfoTreeUI getInfoTreeUI(){
		return _infoTreeUI;
	}
	
	/**
	 * Sets the InfoTreeUI
	 * 
	 * @param inInfoTreeUI The new InfoTreeUI
	 */
	public void setInfoTreeUI(OverviewInfoTreeUI inInfoTreeUI){
		_rightPane.remove(_infoTreeUI);
		_infoTreeUI = inInfoTreeUI;
		_rightPane.add(_infoTreeUI);
		_mainSplitPane.setDividerLocation(_mainSplitPane.getDividerLocation());
	}

	/**
	 * Updates the title of the overview frame to the title stored in the document UI.
	 * Normally called by DocumentInfo.
	 */
	public void updateTitle() {
		String title = _overview.getDocInfo().getName();
		if (title == null || title.equals(""))
			title = "Untitled";

		setTitle("EASIK - " + title);
	}

	/**
	 * Creates the popup menu
	 */
	private void buildPopupMenu() {
		_canvasPopupMenu.add(_AddSketchPopItem = new JMenuItem(new NewSketchAction(_popupPosition, _overview)));
		_canvasPopupMenu.add(_ImportSketchPopItem = new JMenuItem(new ImportSketchAction(_popupPosition, _overview)));

		_sketchPopupMenu.add(_OpenSketchPopDefItem = new JMenuItem(new OpenSketchAction(_overview, true)));
		_sketchPopupMenu.add(_OpenSketchManipulatePopDefItem = new JMenuItem(new OpenSketchDataAction(_overview, true)));
		_sketchPopupMenu.add(_OpenSketchPopItem = new JMenuItem(new OpenSketchAction(_overview)));
		_sketchPopupMenu.add(_OpenSketchManipulatePopItem = new JMenuItem(new OpenSketchDataAction(_overview)));
		_sketchPopupMenu.addSeparator();
		_sketchPopupMenu.add(_NewViewPopItem = new JMenuItem(new NewViewAction(_overview)));
		_sketchPopupMenu.addSeparator();
		_sketchPopupMenu.add(_RenamePopItem = new JMenuItem(new RenameInOverviewAction(_overview)));
		_sketchPopupMenu.add(_DeleteSketchPopItem = new JMenuItem(new DeleteFromOverviewAction(_overview)));
		_sketchPopupMenu.addSeparator();
		_sketchPopupMenu.add(_ExportSketchPopItem = new JMenuItem(new SketchExportAction(this))); 
		_sketchPopupMenu.add(_ExportSketchSQLPopItem = new JMenuItem(new ExportSQLAction(this, null, false)));

		_viewPopupMenu.add(_OpenViewPopItem = new JMenuItem(new OpenViewAction(_overview)));
		_viewPopupMenu.add(_RenameViewPopItem = new JMenuItem(new RenameInOverviewAction(_overview)));
		_viewPopupMenu.add(_DeleteViewPopItem = new JMenuItem(new DeleteFromOverviewAction(_overview)));

		_mixedPopupMenu.add(_DeleteMixedPopItem = new JMenuItem(new DeleteFromOverviewAction(_overview)));
		
		_overview.addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent evt) {
				mouseReleased(evt);
			}
			public void mouseReleased(MouseEvent evt) {
				if (evt.isPopupTrigger()) {
					Object rightClickedOn = _overview.getFirstCellForLocation(evt.getX(),evt.getY());
					// When we have a right click, we check to se if it is part of the selection. If 
					// it isn't, we change the selection to include only that which was
					// under the click, otherwise we'll trigger the popup for the current selection
					if ( rightClickedOn != null){
						Object[] selection = _overview.getSelectionCells();
						if(!Arrays.asList(selection).contains(rightClickedOn)){
							_overview.setSelectionCell(_overview.getFirstCellForLocation(evt.getX(),evt.getY()));
						}
					}
					else{
						_overview.setSelectionCells(new Object[0]);
					}

					_popupPosition.setLocation(evt.getX(), evt.getY());
					JPopupMenu menu = setPopMenu();
					if (menu != null)
						menu.show(evt.getComponent(), evt.getX(), evt.getY());
				}
			}
		});
	}

	private JPopupMenu setPopMenu(){
		Object[] currentSelection = _overview.getSelectionCells();

		if (currentSelection.length == 0) {
			setCanvasPopItems();
			return _canvasPopupMenu;
		}

		// check for mixed selection
		Object selected = currentSelection[0];
		for (Object o: currentSelection) {
			if (o.getClass() != selected.getClass()) {
				setMixedPopItems();
				return _mixedPopupMenu;
			}
		}

		if (selected instanceof SketchNode) {
			setSketchPopItems();
			return _sketchPopupMenu;
		}
		else if (selected instanceof ViewNode){
			setViewPopItems();
			return _viewPopupMenu;
		}
		return null;
	}

	public void setCanvasPopItems(){
		_AddSketchPopItem.setEnabled(true);
		_ImportSketchPopItem.setEnabled(true);
	}

	public void setSketchPopItems(){
		//dissable all elements
		for(Component c : _sketchPopupMenu.getComponents()){
			c.setEnabled(false);
		}
		_OpenSketchManipulatePopDefItem.setVisible(false);
		_OpenSketchManipulatePopItem.setVisible(true);
		_OpenSketchPopDefItem.setVisible(false);
		_OpenSketchPopItem.setVisible(true);

		// We always want to be able to delete
		_DeleteSketchPopItem.setEnabled(true);
		Object[] currentSelection = _overview.getSelectionCells();
		if(currentSelection.length == 1){
			_RenamePopItem.setEnabled(true);

			if ( ((SketchNode)currentSelection[0]).getFrame().getMode() == SketchFrame.Mode.EDIT )
				_NewViewPopItem.setEnabled(true);

			_ExportSketchPopItem.setEnabled(true);

			// Enable sketch export to database
			_ExportSketchSQLPopItem.setEnabled(true);
			// The the export item to work with current selection
			_sketchPopupMenu.remove(_ExportSketchSQLPopItem);
			_ExportSketchSQLPopItem = new JMenuItem(new ExportSQLAction(this, ((SketchNode) currentSelection[0]).getFrame().getSketch(), false));
			_sketchPopupMenu.add(_ExportSketchSQLPopItem 
									= new JMenuItem(new ExportSQLAction(this, ((SketchNode) currentSelection[0]).getFrame().getSketch(), false)));


			if (((SketchNode) currentSelection[0]).getFrame().getSketch().isSynced()) {
				_OpenSketchManipulatePopDefItem.setEnabled(true);
				_OpenSketchManipulatePopDefItem.setVisible(true);
				_OpenSketchManipulatePopDefItem.setArmed(true);
				_OpenSketchManipulatePopItem.setVisible(false);
				_OpenSketchPopItem.setEnabled(true);

			}
			else {
				_OpenSketchPopDefItem.setEnabled(true);
				_OpenSketchPopDefItem.setVisible(true);
				_OpenSketchPopDefItem.setArmed(true);
				_OpenSketchPopItem.setVisible(false);
				_OpenSketchManipulatePopItem.setEnabled(true);
			}
		}
	}

	public void setMixedPopItems(){
		//dissable all elements
		for(Component c : _mixedPopupMenu.getComponents()){
			c.setEnabled(false);
		}
		Object[] currentSelection = _overview.getSelectionCells();

		// We always want delete
		_DeleteMixedPopItem.setEnabled(true);
	}

	public void setViewPopItems(){
		//dissable all elements
		for(Component c : _viewPopupMenu.getComponents()){
			c.setEnabled(false);
		}
		Object[] currentSelection = _overview.getSelectionCells();

		// We always want delete
		_DeleteViewPopItem.setEnabled(true);

		if (currentSelection.length == 1) {
			_RenameViewPopItem.setEnabled(true);
			_OpenViewPopItem.setEnabled(true);
		}

	}

	/**
	 * Builds the menu
	 */
	private void buildMenu() {
		JMenuBar mainMenu;
		JMenu menuFile, menuEdit, menuAction, menuHelp;

		mainMenu = new JMenuBar();

		//Make the File Menu
		menuFile = new JMenu("File");
		addMenuItem(menuFile, new FileNewOverviewAction(this), KeyEvent.VK_N);
		addMenuItem(menuFile, new FileOpenAction(this), KeyEvent.VK_O);
		menuFile.addSeparator();
		addMenuItem(menuFile, new FileSaveAction(this), KeyEvent.VK_S);
		addMenuItem(menuFile, new FileSaveAsAction(this), KeyEvent.VK_S, InputEvent.SHIFT_DOWN_MASK);
		menuFile.addSeparator();
		_recentFilesMenu = new JMenu("Open recent...");
		menuFile.add(_recentFilesMenu);
		updateRecentFilesMenu();

		menuFile.addSeparator();
		addMenuItem(menuFile, new OverviewDocumentInfoAction(this), KeyEvent.VK_I);

		menuFile.addSeparator();
		addMenuItem(menuFile, new FileQuitAction(this), KeyEvent.VK_Q);
		mainMenu.add(menuFile);
		menuFile.setMnemonic(KeyEvent.VK_F);

		menuEdit = new JMenu("Edit");
		menuEdit.add(new JMenuItem(new NewSketchAction(null, _overview)));
		menuEdit.add(new JMenuItem(new ImportSketchAction(null, _overview)));
		menuEdit.addSeparator();
		menuEdit.add(_NewViewMenuItem = new JMenuItem(new NewViewAction(_overview)));
		mainMenu.add(menuEdit);
		menuEdit.addMenuListener(new MenuListener() {
			public void menuSelected(MenuEvent e1) {
				Object[] selection = _overview.getSelectionCells();
				_NewViewMenuItem.setEnabled( selection.length == 1 && selection[0] instanceof SketchNode);
			}
			public void menuDeselected(MenuEvent e1){}
			public void menuCanceled(MenuEvent e1){}
		});
		menuEdit.addSeparator();
		addMenuItem(menuEdit, _ProgramSettingsItem = new JMenuItem("Preferences"), KeyEvent.VK_E); // If you change this title, update OSX.java
		_ProgramSettingsItem.setToolTipText("Set Global Easik Properties");
		_ProgramSettingsItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e1) {
				ProgramSettingsUI myUI = new ProgramSettingsUI(ApplicationFrame.this);
				myUI.showDialog();
			}
		});
		menuEdit.setMnemonic(KeyEvent.VK_E);

		//Create help menu
		menuHelp = new JMenu("Help");
		mainMenu.add(menuHelp);
		menuHelp.setMnemonic(KeyEvent.VK_H);
		menuHelp.add(new HelpAction());
		menuHelp.add(new AboutAction());

		setJMenuBar(mainMenu);
	}

	/**
     * Updates the recent files menu items
     */
    public void updateRecentFilesMenu() {
        _recentFilesMenu.removeAll();
        for (String file: _settings.getPropertySet("recent_files")) {
            _recentFilesMenu.add(new JMenuItem(new RecentFileAction(file, this)));
        }
    }

	/**
	 * Adds the specified File to the recent files menu, then updates
	 * the recent files menu items to reflect the new list of files.
	 *
	 * @param selFile the file to add to the recent files list
	 */
	public void addRecentFile(File selFile) {
		Easik.getInstance().getSettings().addRecentFile(selFile);
		updateRecentFilesMenu();
	}

	/**
	 * Called by Overview when the overview is set as dirty (or clean).  We
	 * simply set the windowModified property for OS-specific window changes
	 * (for example, the "modified" dot in OS X
	 */
	public void setDirty(boolean dirty) {
		getRootPane().putClientProperty("windowModified", dirty ? Boolean.TRUE : Boolean.FALSE);
	}

	/**
	 * Sets the window adjuster object that will be called whenever a window
	 * with menus is created to adjust the menus in an OS-specific way.  If
	 * null, no menu adjustment is performed.  If non-null, this immediately
	 * calls the adjuster on the main application frame.
	 */
	public void setMenuAdjuster(EasikMenuAdjuster adj) {
		_menuAdj = adj;
		if (_menuAdj != null)
			_menuAdj.adjustMenu(getJMenuBar());
	}

	/**
	 * Returns the current menu adjuster object.
	 */
	public EasikMenuAdjuster getMenuAdjuster() {
		return _menuAdj;
	}
}
