package easik.ui.datamanip;

import javax.swing.*;
import java.awt.Window;

import java.util.*;
import java.sql.*;

import easik.database.DatabaseDriver;
import easik.database.ColumnNaming;
import easik.database.types.*;
import easik.sketch.attribute.EntityAttribute;
import easik.sketch.vertex.EntityNode;
import easik.sketch.edge.SketchEdge;
import easik.view.vertex.QueryNode;

public abstract class DatabaseUtil{

	/**
	 * Pops up a dialog for the user to select records from the table
	 * represented by the given entity node. The primary IDs of the records are
	 * returned.
	 * @param frame the JFrame to attach the dialog to
	 * @param table The entity node representing the table from which we wish to select the records
	 * @return An array of primary IDs representing the records we selected
	 */
	public static int[] selectRowPKs(JFrame frame, EntityNode table){
		DatabaseDriver dbd;
		JTable tableData;
		try{
			dbd = table.getSketch().getDatabaseDriver();
			tableData = getTable(table);
		}
		catch(DatabaseDriver.LoadException e){
			System.err.println("Error in DatabaseUtil: " + e.getMessage());
			return null;
		}
		catch(SQLException e){
			System.err.println("Error in DatabaseUtil: " + e.getMessage());
			return null;
		}

		SelectDataDialog sdd = new SelectDataDialog(frame, table.getName(), tableData, dbd);
		if(sdd.isAccepted()){
			return sdd.getSelectedPKs();
		}
		return new int[0];
	}

	/**
	 * Pops up a dialog for the user to select a single record from the table
	 * represented by the given entity node. The primary ID of the selected
	 * record is returned.
	 * @param frame the JFrame or JDialog to attach the dialog to
	 * @param table The entity node representing the table from which we wish to select the record
	 * @return The primary ID representing the record we selected
	 */
	public static int selectRowPK(Window win, EntityNode table){
		DatabaseDriver dbd;
		JTable tableData;
		try{
			dbd = table.getSketch().getDatabaseDriver();
			tableData = getTable(table);
			tableData.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		}
		catch(DatabaseDriver.LoadException e){
			System.err.println("Error in DatabaseUtil: " + e.getMessage());
			// 0 is an invalid PK
			return 0;
		}
		catch(SQLException e){
			System.err.println("Error in DatabaseUtil: " + e.getMessage());
			// 0 is an invalid PK
			return 0;
		}
		SelectDataDialog sdd = win instanceof JDialog
			? new SelectDataDialog((JDialog) win, table.getName(), tableData, dbd)
			: new SelectDataDialog((JFrame) win, table.getName(), tableData, dbd);
		if(sdd.isAccepted()){
			// recall that we set our table for singleton selection
			int[] selected = sdd.getSelectedPKs();
			return selected[0];
		}
		// 0 is an invalid PK
		return 0;

	}

	/**
	 * Returns an array of the primary keys found in a given table. An array of
	 * size 0 is returned in the event of a database access error.
	 * @param table The entity node representing the table from which we wish
	 * to get the primary IDs
	 * @return The primary IDs of all records found in the given table
	 */
	public static int[] getPKs(EntityNode table){
		DatabaseDriver dbd;
		ColumnNaming cn;
		ResultSet result;
		try{
			dbd = table.getSketch().getDatabaseDriver();
			cn = new ColumnNaming(dbd);
			result = dbd.executeQuery("SELECT * FROM " + table.getName());

			result.last(); // move to last row to get row count
			int[] pKs = new int[result.getRow()];

			// move back to first row
			result.beforeFirst();
			result.next();

			String pIdName = cn.tablePK(table);
			for(int i=0; i<pKs.length; i++){
				pKs[i] = result.getInt(pIdName);
				result.next();
			}

			return pKs;
		}
		catch(DatabaseDriver.LoadException e){
			System.err.println("Error in DatabaseUtil: " + e.getMessage());
			return new int[0];
		}
		catch(SQLException e){
			System.err.println("Error in DatabaseUtil: " + e.getMessage());
			return new int[0];
		}
	}

	/**
	 * Gets the contents of a given table, and returns them as a JTable. The
	 * first column is the record's primary key, and the remaining columns are
	 * layed out in the order EntityNode.getEntityAttributes() is stored.
	 * @param table The EntityNode representing the table we wish to view
	 * @param dbd The database driver defining formats we should use in creating this table
	 * @return The contents of the given table wrapped as a JTable
	 */
	public static JTable getTable(EntityNode table)throws SQLException, DatabaseDriver.LoadException{
		DatabaseDriver dbd;
		ColumnNaming cn;
		Vector<Object[]> data;
		String[] columnNames;

		int cols;

		// Get column names of "shadow edges". We will ignore these columns.
		HashSet<SketchEdge> shadowEdges = new HashSet<SketchEdge>(table.getShadowEdges());
		HashSet<String> shadowColumns = new HashSet<String>(shadowEdges.size());
		for(SketchEdge edge : shadowEdges){
			shadowColumns.add(edge.getName());
		}

		ResultSet result;

		dbd = table.getSketch().getDatabaseDriver();
		cn = new ColumnNaming(dbd);
		String mess = "SELECT * FROM " + dbd.quoteId(table.getName());
		result = dbd.executeQuery(mess);
		ResultSetMetaData rsmd = result.getMetaData();

		// build 2D array of data
		cols = rsmd.getColumnCount() - shadowEdges.size(); 

		data = new Vector();
		columnNames = new String[cols];

		int currRow, i;

		// Populate array of column names
		for(i=0; i<cols; i++){
			String columnName = rsmd.getColumnName(i+1);
			if ( ! (shadowColumns.contains(columnName)) ) 
				columnNames[i] = columnName;
		}

		while(result.next()){
			currRow = result.getRow();
			// if we have no rows, break
			if( currRow == 0 )
				break;
			Object[] row = new Object[cols];
			for(i=0; i<cols; i++){
				row[i] = result.getString(columnNames[i]);
			}
			data.add(row);
		}

		Object[][] dataObj = data.toArray(new Object[0][0]);
		return new JTable(dataObj, columnNames);
	}

	/**
	 * Returns a Map of column names to their value for a given table/primary key pair. If no record
	 * exists in the table with the given primary key, we return an empty Map. Can throw a 
	 * DatabaseDriver.LoadException or SQLException in the event that a database driver can not be acquired.
	 * @param table The table from which we get the requested row.
	 * @param pk The primary id of the row which we want
	 * @return A mapping of column names to their corresponding value for the given table/id pair. If no
	 * such row exists, returns an empty Map.
	 */
	public static Map<String, String> getRecord(EntityNode table, int pk){
		DatabaseDriver dbd;
		ColumnNaming cn;
		HashMap<String, String> data = new HashMap();

		ResultSet result;
		try{
			dbd = table.getSketch().getDatabaseDriver();
			cn = new ColumnNaming(dbd);
			String select = "SELECT * FROM "
					+ dbd.quoteId(table.getName())
					+ " WHERE "
					+ cn.tablePK(table)
					+ " = "
					+ pk;
			result = dbd.executeQuery(select);
			ResultSetMetaData rsmd = result.getMetaData();

			result.next();

			for( int i=1; i<=rsmd.getColumnCount(); i++){
				data.put(dbd.quoteId(rsmd.getColumnName(i)), result.getString(i));
			}
		}
		catch(DatabaseDriver.LoadException e){
			JOptionPane.showMessageDialog(table.getSketch().getFrame(), "Could not get record: " + e.getMessage());
		}
		catch(SQLException e){
			JOptionPane.showMessageDialog(table.getSketch().getFrame(), "Could not get record: " + e.getMessage());
		}
		return data;
	}

	public static void displayQueryNode(QueryNode node, JFrame parent) {
		DatabaseDriver dbd;
		ResultSet result;
		ResultSetMetaData rsmd;
		Vector<Object[]> data;
		JTable asTable;
		String[] columnNames;
		int cols;

		try{ 
			dbd = node.getView().getSketch().getDatabaseDriver(); 
			result = dbd.executeQuery(node.getQuery());

			rsmd = result.getMetaData();

			cols = rsmd.getColumnCount();

			// build 2D array of data
			data = new Vector();
			columnNames = new String[cols];

			int currRow, i;

			// Populate array of column names
			for(i=0; i<cols; i++){
				columnNames[i] = rsmd.getColumnName( i + 1 );
			}

			while(result.next()){
				currRow = result.getRow();
				// if we have no rows, break
				if( currRow == 0 )
					break;
				Object[] row = new Object[cols];
				for(i=0; i<cols; i++){
					row[i] = result.getString(columnNames[i]);
				}
				data.add(row);
			}

			Object[][] dataObj = data.toArray(new Object[0][0]);
			asTable = new JTable(dataObj, columnNames);

			new SelectDataDialog(parent, node.getName(), asTable, dbd);
		}
		catch(DatabaseDriver.LoadException e){ 
			JOptionPane.showMessageDialog(
					node.getView().getSketch().getFrame(), "Could not get record: " + e.getMessage()); 
			return;
		}
		catch(SQLException e){ 
			JOptionPane.showMessageDialog(
					node.getView().getSketch().getFrame(), "Could not get record: " + e.getMessage()); 
			return;
		}
	}
}
