package easik.ui.datamanip;

import java.util.List;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JOptionPane;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.FlowLayout;
import java.awt.GridLayout;

import java.util.*;

import java.sql.*;

import easik.database.DatabaseDriver;
import easik.database.types.*;
import easik.sketch.vertex.EntityNode;
import easik.database.types.Type;
import easik.ui.*;

/**
 * Sets up a dialog for the user to enter data. The display will include a label indicating each field's corresponding column,
 * and the datatype expected for the entry. Default values for fields can be included in the form of a map from column name
 * to the default text can be included. 
 */

public class RowEntryDialog extends OptionsDialog{

	private JFrame _parent;

	/** Options shown by this dialog */
	private LinkedList<Option> options;

	/** Input from the user in the form of a ColumnEntry object set */
	private HashSet<ColumnEntry> input;

	/** A map of column names to their correspoinding data type */
	private HashMap<String, Type> attsToTypes;

	/** A map foreign key name to a JTable holding the data of the table to which the foreign key points */
	private LinkedHashMap<String, EntityNode> foreignKeys;

	/** Set of DialogRow objects which are used by the verify method to grab input */
	private HashSet<DialogRow> dialogRows;

	/** The driver for database communication */
	private DatabaseDriver dbd;

	/** Map of default column data for text fields, indexed by column name */
	private HashMap<String, String> defaults;

	/**
	 * Sets up dialog for inserting data into a table
	 *
	 * @param parent The parent frame of the modal dialog
	 * @param name The name of the table into which we are inserting data
	 * @param inAttToType A map of attribute (column) names, to the type of input they expect
	 * @param fKeys A map of column names which represent foreign keys to the EntityNode to which it points
	 * @param defaults Default values for entry fields in the form of a map from column name to default text
	 */
	public RowEntryDialog(JFrame parent, 
			DatabaseDriver inDBD, 
			String title, 
			Map<String, Type> inAttToType, 
			Map<String, EntityNode> fKeys,
			Map<String, String> defs) {
		super(parent, title);

		attsToTypes = new HashMap(inAttToType);
		foreignKeys = new LinkedHashMap(fKeys);
		defaults = new HashMap(defs);
		dialogRows = new HashSet();

		dbd = inDBD;
		
		_parent = parent;
		setSize(465, 400);
		this.showDialog();
	}

	public RowEntryDialog(JFrame parent, 
			DatabaseDriver inDBD, 
			String title, 
			Map<String, Type> inAttToType, 
			Map<String, EntityNode> fKeys) {
		this(parent, inDBD, title, inAttToType, fKeys, new HashMap<String, String>());
	}

	/**
	 * Called when the user clicks the OK button. If the input for a given field is deemed valid, it is added
	 * to the input map. If any input is invalid. The whole input map is set to null and the method returns false.
	 *
	 * @return true if all inputs check out, false otherwise.
	 */
	public boolean verify() {
		input = new HashSet();
		JComponent inputComponent;
		String fieldInput;
		String column;


		for(DialogRow row : dialogRows){
			column = row.columnName;
			inputComponent = row.input;

			// verify "special" types - we selected value from a popup, so know it to be valid
			// If this column is a foreign key, picked from table -> Assume valid and is an INT
			if(foreignKeys.containsKey(column)){
				fieldInput = ((JTextField)inputComponent).getText();
				input.add(new ColumnEntry(column, fieldInput, new Int()));
			}
			else if(inputComponent instanceof JDateChooser){
				// grab input from date chooser - null -> bad input
				String date = Long.toString(((JDateChooser)inputComponent).getDate().getTime());
				if(date == null){// bad input
					JOptionPane.showMessageDialog(
						this,
						"<html>Invalid date selected for <b>" + column + "</b>",
						column + ": Invalid input",
						JOptionPane.ERROR_MESSAGE);
					input = null;
					return false;
				}
				input.add(new ColumnEntry(column, date, row.type));
			}
			else if(inputComponent instanceof TimeChooser){
				TimeChooser t = (TimeChooser)inputComponent;
				input.add(new ColumnEntry(column, Long.toString(t.getMillis()), row.type));
			}
			else if(inputComponent instanceof DateTimeChooser){
				DateTimeChooser dtc = (DateTimeChooser)inputComponent;
				java.util.Date date = dtc.dateChooser.getDate();
				if(date == null){// bad input
					JOptionPane.showMessageDialog(
						this,
						"<html>Invalid date/time for <b>" + column + "</b>",
						column + ": Invalid input",
						JOptionPane.ERROR_MESSAGE);
					input = null;
					return false;
				}
				long dateTime = dtc.getMillis();
				input.add(new ColumnEntry(column, Long.toString(dateTime), row.type));
			}
			else{

				fieldInput = ((JTextField)inputComponent).getText();

				// Allow "" (treated as NULL), or ask the type expected to validate the input
				if( fieldInput.equals("") || row.type.verifyInput(fieldInput) ){
					input.add(new ColumnEntry(column, fieldInput, row.type));
				}
				else{// otherwise we have bad input
					JOptionPane.showMessageDialog(
						this,
						"<html>Invalid input '" + fieldInput + "' for <b>" + column + "</b> (" + row.type + ")",
						column + ": Invalid input",
						JOptionPane.ERROR_MESSAGE);
					input = null;
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Returns a map of column names to its input.
	 * @return The map of column names to input if verify has benn called and returned true, otherwise null.
	 */
	public Set<ColumnEntry> getInput(){
		return input;
	}

	public List<Option> getOptions(){
		options = new LinkedList();

		if (attsToTypes.isEmpty() && foreignKeys.isEmpty()){
			options.add(new Option("<html><i>(No attributes or outgoing edges)</i><br>Press OK to add a new row."));
			setSize(300, 145);
			return options;
		}

		for(String name : attsToTypes.keySet()){
			
			Type type = attsToTypes.get(name);
			JComponent c;

			// set up special case types where we restrict user input to a selection, and
			// add a corresponding DialogRow to our set of DialogRows
			if(type instanceof easik.database.types.Date){
				c = JUtils.fixHeight(new JDateChooser());
			}
			else if(type instanceof easik.database.types.Time){
				c = JUtils.fixHeight(new TimeChooser());
			}
			else if(type instanceof easik.database.types.Timestamp){
				c = JUtils.fixHeight(new DateTimeChooser());
			}
			// else allow any input
			else{
				c = JUtils.textField("");
				c.setToolTipText(type.toString());
				if( defaults.keySet().contains(name)){
					((JTextField)c).setText(defaults.get(name));
				}
			}
			options.add(new Option("<html><b>"+name+"</b>", c));
			dialogRows.add( new DialogRow(c, name, type) );
		}

		for(String fkName : foreignKeys.keySet()){
			// Grab target table name for "type"
			String targetType = foreignKeys.get(fkName).getName();

			JTextField idDisplay = JUtils.textField(5);
			if( defaults.keySet().contains(fkName)){
				idDisplay.setText(defaults.get(fkName));
			}
			idDisplay.setEditable(false);

			JButton b = new JButton("Choose " + targetType);
			b.addActionListener(new ForeignKeyListener(foreignKeys.get(fkName), idDisplay));

			options.add(new Option("<html><b>"+fkName+"</b>", idDisplay, b));
			dialogRows.add( new DialogRow(idDisplay, fkName, new Int()));
		}

		if(options.size() < 5)
			setSize(465, 275);
		return options;
	}

	/**
	 * Listens on the foreign key-selecting button and opens a singleton record-selection dialog. The primary
	 * ID of the selected record is then set in the _display text field for the user.
	 */
	private class ForeignKeyListener implements ActionListener{
		private EntityNode _table;
		private JTextField _display;

		public ForeignKeyListener(EntityNode inTable, JTextField display){
			super();
			_table = inTable;
			_display = display;
		}
		// Fired when the user clicks on a button in a field containing a foreign key
		public void actionPerformed(ActionEvent e){
			int selected = DatabaseUtil.selectRowPK(RowEntryDialog.this, _table);
			// if 0, there were no valid foreign keys to choose
			if(selected == 0){
				return;
			}
			_display.setText(Integer.toString(selected));
		}
	}
	/**
	 * JPanel for entering time via JSpinners.
	 */
	private class TimeChooser extends JPanel{
		private JSpinner h, m, s, ms;

		/**
		 * Sets up time chooser with default values for the spinners. Bad defaults get set to 0.
		 */
		private TimeChooser(int hr, int min, int sec, int msec){
			if(hr>23)
				hr = 0;
			if(min>59)
				min = 0;
			if(sec>59)
				sec = 0;
			if(msec>999)
				msec = 0;

			setLayout(new FlowLayout(FlowLayout.LEFT, 2, 0));
			
			this.add(h = new JSpinner(new SpinnerNumberModel(hr, 0, 23, 1)));
			this.add(new JLabel(":"));
			this.add(m = new JSpinner(new SpinnerNumberModel(min, 0, 59, 1)));
			this.add(new JLabel(":"));
			this.add(s = new JSpinner(new SpinnerNumberModel(sec, 0, 59, 1)));
			this.add(new JLabel("."));
			this.add(ms = new JSpinner(new SpinnerNumberModel(msec, 0, 999, 1)));

			h.setToolTipText("h");
			m.setToolTipText("m");
			s.setToolTipText("s");
			ms.setToolTipText("ms");

		}

		/**
		 * Default constructor. Initializes all spinners to 0.
		 */
		private TimeChooser(){
			this(0, 0, 0, 0);
		}

		/**
		 * Returns the miliseconds represented by this time chooser, corrected via Java's Calendar object
		 * so that when sent to an SQL server, the value interpreted is the same as the user sees on the JSpinners
		 */
		private long getMillis(){
			Calendar c = Calendar.getInstance();
			c.clear();

			c.set(Calendar.HOUR_OF_DAY, ((Integer)h.getValue()));
			c.set(Calendar.MINUTE, (Integer)m.getValue());
			c.set(Calendar.SECOND, (Integer)s.getValue());
			c.set(Calendar.MILLISECOND, (Integer)ms.getValue());

			return c.getTimeInMillis();
		}
	}

	private class JDateChooser extends com.toedter.calendar.JDateChooser{
		/**
		 * Returns the selected date, with time components set to the values selected in the TimeChooser object
		 * @param tc The TimeChooser object who's time fields are used to set times in our date
		 * @return The Date object selected with hours, minutes, seconds and milliseconds set to reflect our TimeChooser
		 */
		public java.util.Date getDate(TimeChooser tc){
			java.util.Date date = super.getDate();

			Calendar c = Calendar.getInstance();
			c.clear();
			c.setTime(date);

			c.set(Calendar.HOUR_OF_DAY, (Integer)tc.h.getValue());
			c.set(Calendar.MINUTE, (Integer)tc.m.getValue());
			c.set(Calendar.SECOND, (Integer)tc.s.getValue());
			c.set(Calendar.MILLISECOND, (Integer)tc.ms.getValue());

			return c.getTime();
		}

		/**
		 * Returns the date selected, with time fields set to midnight
		 */
		public java.util.Date getDate(){
			return getDate(new TimeChooser());
		}
	}

	/**
	 * JComponent for entering a date and time. Constains JSpinners for hour:min:sec and a JDateChooser.
	 */
	private class DateTimeChooser extends JPanel{
		private TimeChooser timeChooser;
		private JDateChooser dateChooser;

		private DateTimeChooser(){
			this.setLayout(new GridLayout(2,1));
			this.add(dateChooser = new JDateChooser());
			this.add(timeChooser = new TimeChooser());
			JUtils.borderless(this);
		}

		private long getMillis(){
			return dateChooser.getDate(timeChooser).getTime();
		}
	}

	/**
	 * Bundle for row in dialog.
	 */
	private class DialogRow{
		private JComponent input;
		private String columnName;
		private Type type;

		private DialogRow(JComponent in, String cn, Type t){
			input = in;
			columnName = cn;
			type = t;
		}
	}
}
