package easik.ui.datamanip;

import java.sql.*;
import java.util.*;
import javax.swing.JTable;
import javax.swing.JOptionPane;

import easik.EasikTools;
import easik.database.*;
import easik.database.types.*;
import easik.sketch.Sketch;
import easik.sketch.attribute.EntityAttribute;
import easik.sketch.edge.*;
import easik.sketch.vertex.EntityNode;
import easik.sketch.constraint.*;
import easik.sketch.path.*;
import easik.ui.datamanip.*;


/**
 * In charge of monitoring update statements sent to server. Ensures constraints
 * are maintained, and appropriate error messages are sent to the user.
 */
public class UpdateMonitor{

	/** The sketch within which our editing takes place */
	private Sketch _theSketch;

	/** The database driver for this update monitor */
	private DatabaseDriver dbd;
	/** A column naming object for this update monitor */
	private ColumnNaming cn;

	public UpdateMonitor(Sketch inSketch, DatabaseDriver inDbd){
		_theSketch = inSketch;
		dbd = inDbd;
		cn = new ColumnNaming(dbd);
	}

	/**
	 * Pops up a dialog for the user to select a record from the given table. Once selected, the
	 * table and primary ID is handed off to updateRow(EntityNode table, int pk).
	 * @param table The entity node representing the table we wish to update.
	 * @return Success of update.
	 */
	public boolean updateRow(EntityNode table){
		int pk = DatabaseUtil.selectRowPK(table.getSketch().getFrame(), table);
		if(pk > 0){
			return updateRow(table, pk);
		}
		return false;
	}

	/**
	 * Pops up a row entry dialog for the given table. The default field values will be the values stored
	 * in the row of the table who's primary ID matches the one given. Some cases call for use to restrict
	 * the update of foreign keys. Those cases are:
	 * 		a) A pullback
	 * 		b) The summands of a sum constrain
	 * 		c) The product of a product constraint
	 * If the table is in a constraint and the user wishes to break it the update is aborted leaving the
	 * old values in the record.
	 * Note: There are currently no checks to ensure that pk matches a record in our table.
	 * @param table The EntityNode representing the table we wish to update.
	 * @param pk The primary ID of the row whos data will be used to set defaults for our row entry dialog.
	 * @return The Success of the update.
	 */
	public boolean updateRow(EntityNode table, int pk){
		DialogOptions dOpts = getDialogOptions(table);
		try{
			// Remove option to update foreign keys restricted by constraints
			for(Constraint c : table.getConstraints()){
				if ( c instanceof SumConstraint && c.getPaths().get(0).getCoDomain() == table 
				  || c instanceof ProductConstraint && c.getPaths().get(0).getDomain() == table
				  || c instanceof PullbackConstraint && ((PullbackConstraint)c).getSource() == table
				   ){
					for(SketchEdge se : table.getOutgoingEdges()){
						if(c.hasEdge(se)){
							dOpts.fKeys.remove(cn.tableFK(se));
						}
					}
				   }
			}
			// Display dialog for user with defaults as old record values
			RowEntryDialog red = new RowEntryDialog(_theSketch.getFrame(), 
					dbd, 
					"Update " + table.getName() + ": " + cn.tablePK(table) + " = " + pk,
					dOpts.attToType,
					dOpts.fKeys,
					DatabaseUtil.getRecord(table, pk));
			if(red.isAccepted()){
				LinkedHashSet<ColumnEntry> input = new LinkedHashSet(red.getInput());

				// if we do not have any new input, no need to execute an update
				if( input.isEmpty() ){
					return true;
				}

				StringBuilder sb = new StringBuilder("UPDATE " + 
						table.getName() + 
						" SET ");
				for( ColumnEntry ce : input ){
					sb.append(dbd.quoteId(ce.getColumnName()) + "=" + "?,");
				}
				// remove last comma
				sb.delete(sb.length()-1, sb.length());

				sb.append(" WHERE " + cn.tablePK(table) + "=" + pk);

				dbd.executePreparedUpdate(sb.toString(), input);
				return true;
			}
			return false;
		}
		catch(SQLException e){
			JOptionPane.showMessageDialog(table.getSketch().getFrame(), "Could not update record: " + e.getMessage());
			return false;
		}
	}

	/**
	 * Trys to delete from a given table. If the action will break a constraint, it is aborted and the
	 * user is notified.
	 * @param table The tabel from which we will attempt the delete
	 * @return The success of the delete
	 */
	public boolean deleteFrom(EntityNode table){

		int[] selectedPKs = DatabaseUtil.selectRowPKs(table.getSketch().getFrame(), table);
		// if there was a selection
		if(selectedPKs.length>0){
			String PKcolumn = cn.tablePK(table);

			StringBuilder sb = new StringBuilder("DELETE FROM "
					+ dbd.quoteId(table.getName())
					+ " WHERE ");

			// populate input set for prepared statement while adding column names
			LinkedHashSet<ColumnEntry> input = new LinkedHashSet();
			for( int pk : selectedPKs ){
				sb.append(PKcolumn + "=? OR ");
				input.add( new ColumnEntry(PKcolumn, Integer.toString(pk), new Int()) );
			}
			// remove last ',OR '
			sb.delete(sb.length()-4, sb.length());

			try{
				dbd.executePreparedUpdate(sb.toString(), input);
			}
			catch(SQLException e){
				JOptionPane.showMessageDialog(null, "Unable to execute DELETE: " + e.getMessage());
			}
		}
		return true;
	}

	/**
	 * Determines if intsertion into a given table requires special handling due to constraints it may be in. As of now,
	 * special cases that may result from being in multiple constraints are not supported.
	 * @param table The table into which we wish to insert data
	 * @return Success of the insertion
	 */
	public boolean insert(EntityNode table){
		DialogOptions dOpts = getDialogOptions(table);
		// a set of column-value pairs of which we wish to force a specific value, leaving the user out
		HashSet<ColumnEntry> forced = new HashSet();

		//FIXME: This is pretty unpredicatable when there is more than one contstraint. Tighten up?
		for( Constraint c : table.getConstraints() ){
			if ( c instanceof SumConstraint ){
				// if table is in the sum's domain, we let out trigger take care of its foreign key,
				// so remove it from the dialog's selection
				for( SketchPath sp : c.getPaths() ) {
					if ( sp.getDomain() == table ) {
						// we force the value 0 to avoid out driver to kick back an error for having a null fKey
						String columnName = cn.tableFK(sp.getFirstEdge());
						dOpts.fKeys.remove(columnName);
						forced.add(new ColumnEntry(columnName, "0", new Int()));
						break;
					}
				}
			}
			if ( c instanceof CommutativeDiagram ){
				// if in the domain of a CD, we must make sure that our paths commute
				if (c.getPaths().get(0).getDomain() == table) {
					JOptionPane.showMessageDialog(
							null,
							"Be sure that the following paths commute:\n" + EasikTools.join("\n", c.getPaths()),
							"Commutative diagram",
							JOptionPane.INFORMATION_MESSAGE);
					try{
						return promptAndInsert(table, dOpts, forced);
					}
					catch(SQLException e){
						JOptionPane.showMessageDialog(
								null,
								"Not all of the following paths commute -- insert aborted!\n" + EasikTools.join("\n", c.getPaths()),
								"Commutative diagram failure",
								JOptionPane.ERROR_MESSAGE);
					}
				}
			}
			if ( c instanceof PullbackConstraint ){
				// if we're not inserting to the target (or source, but that is dissabled at the popup menu level), there is
				// a chance our insert will pull a record back into the source. If this happens, we want to let the user update
				// the new record
				if ( ((PullbackConstraint)c).getTarget() != table){
					EntityNode pullback = ((PullbackConstraint)c).getSource();
					try{
						// get row count pre-insert
						ResultSet result = dbd.executeQuery("SELECT COUNT(*) FROM " + pullback.getName() + " X");
						result.next();
						int preRowCount = result.getInt(1);

						promptAndInsert(table, dOpts);

						// get row count post-insert
						result = dbd.executeQuery("SELECT COUNT(*) FROM " + pullback.getName() + " X");
						result.next();
						int postRowCount = result.getInt(1);

						// if our pullback has more rows after INSERT, update new row (the one with the highest primary ID)
						if(postRowCount > preRowCount){
							result = dbd.executeQuery("SELECT MAX(" + cn.tablePK(pullback) + ") FROM " + pullback.getName() + " X");
							result.next();
							int pk = result.getInt(1);

							if ( JOptionPane.showConfirmDialog(null, "New record in pullback table '" 
									+ pullback.getName() 
									+ "'. Enter column data?",
									"Insert column data?",
									JOptionPane.YES_NO_OPTION) == 0)
								updateRow(pullback, pk);
						}
						return true;
					}
					catch(SQLException e){
						JOptionPane.showMessageDialog(null, "Could not execute update: " + e.getMessage());
					}
				}
			}
			if ( c instanceof ProductConstraint ){
				// if we're inserting into a factor, we want to allow the user to set values for each resulting record
				// automatically inserting into the product.
				for( SketchPath sp : c.getPaths()){
					if(sp.getCoDomain() == table){
						EntityNode product = sp.getDomain();

						try{
							promptAndInsert(table, dOpts);

							// get the new records from the product. They are any record who's fk to our INSERT factor matches
							// the primary id of the last insert
							ResultSet result = dbd.executeQuery("SELECT MAX(" + cn.tablePK(table) + ") FROM " + table.getName() + " X");
							result.next();
							int newPK = result.getInt(1);


							result = dbd.executeQuery("SELECT * FROM " 
									+ product.getName() 
									+ " WHERE " 
									+ cn.tableFK(sp.getFirstEdge())
									+ " = "
									+ newPK);

							// get count of new rows as result of INSERT
							result.last();
							int newRows = result.getRow();
							result.beforeFirst();


							if ( newRows > 0 && 
									JOptionPane.showConfirmDialog(null, newRows 
										+ " new rows in product table '"
										+ product.getName()
										+ "'. Insert column data?",
										"Insert column data?",
										JOptionPane.YES_NO_OPTION) == 0)
								while(result.next())
									updateRow(product, result.getInt(1));
						}
						catch(SQLException e){
							JOptionPane.showMessageDialog(null, e.getMessage());
						}

						return true;

					}
				}
			}
		}

		try{
			return promptAndInsert(table, dOpts, forced);
		}
		catch(SQLException e){
			JOptionPane.showMessageDialog(null, "Could not execute update: " + e.getMessage());
			return false;
		}
	}

	/**
	 * Gets the options needed for open a dialog for row insertion: map of attribute names to their type, and
	 * a map of foreign key names to a JTable of data they point to.
	 * @param table The node representing the table who's information we want
	 * @return Wrapper class holding the maps needed for insertion dialog
	 */
	private DialogOptions getDialogOptions(EntityNode table){
		HashSet<Constraint> constraints = new HashSet(table.getConstraints());
		HashMap<String, Type> attToType = new HashMap();
		LinkedHashMap<String, EntityNode> fKeys = new LinkedHashMap();

		// find attributes, and map to their Type
		for (EntityAttribute ea: table.getEntityAttributes())
			attToType.put(ea.getName(), ea.getType());
		// find all foreign keys, and add to foreign key set
		for (SketchEdge ske: table.getOutgoingEdges())
			fKeys.put(cn.tableFK(ske), ske.getTargetEntity());
		// find shadow foreign keys (used by sum constraints to solve parent insertion issues)
		for (SketchEdge shadowEdge: table.getShadowEdges())
			fKeys.put(cn.tableFK(shadowEdge), shadowEdge.getTargetEntity());

		return new DialogOptions(attToType, fKeys);
	}

	/**
	 * Initializes a data insertion dialog and waits for the user to insert and accept the data. Once Accepted,
	 * forms an input statement and a map of input to its type which is handed to DatabaseDriver.executePreparedUpdate(..).
	 * @param table the EntityNode
	 * @param dOpts A DialogOptions object containing maps as specified in easik.ui.datamanip.RowEntryDialog
	 * @param forced A set of ColumnEntry objects that will be used in the generated INSERT statement which the user
	 *    has no control over (i.e. the columns hadn't not shown up in the entry dialog, but we still wish to include them). 
	 *    No column name in this set should match one in dOpts. Should this happen, is is undetermined which which of the
	 *    two entries will appear first in the input, and therefore the result after INSERT is undetermined.
	 * @return Success of insert
	 */
	private boolean promptAndInsert(EntityNode table, DialogOptions dOpts, Set<ColumnEntry> forced) throws SQLException{
		String tableName = table.getName();
		RowEntryDialog red = new RowEntryDialog(table.getSketch().getFrame(), dbd, "Add row to table: " + tableName, dOpts.attToType, dOpts.fKeys);

		if(red.isAccepted()){
			LinkedHashSet<ColumnEntry> input = new LinkedHashSet(red.getInput());
			input.addAll(forced);

			StringBuilder sb = new StringBuilder("INSERT INTO " + dbd.quoteId(tableName) + " ");

			LinkedList<String> colNames = new LinkedList<String>(), values = new LinkedList<String>();

			for (ColumnEntry entry : input) {
				colNames.add(dbd.quoteId(entry.getColumnName()));
				values.add("?");
			}
			if (colNames.isEmpty())
				sb.append(dbd.emptyInsertClause());
			else
				sb.append("(" + EasikTools.join(", ", colNames) + ") VALUES (" + EasikTools.join(", ", values) + ")");

			dbd.executePreparedUpdate(sb.toString(), input);

			return true;
		}
		return false;
	}

	/**
	 * Calls promptAndInsert(EntityNode table, DialogOptions dOpts, Set<ColumnEntry> forced) with an empty set of forced
	 * values.
	 */
	private boolean promptAndInsert(EntityNode table, DialogOptions dOpts) throws SQLException{
		return promptAndInsert(table, dOpts, new HashSet<ColumnEntry>());
	}

	private class DialogOptions{
		private HashMap<String, Type> attToType;
		private LinkedHashMap<String, EntityNode> fKeys;

		private DialogOptions(Map<String, Type> inAttToType, Map<String, EntityNode> inFKeys){
			attToType = new HashMap(inAttToType);
			fKeys = new LinkedHashMap(inFKeys);
		}
	}
}
