package easik.ui.menu.popup;

import easik.Easik;
import easik.database.DatabaseDriver;
import easik.ui.JUtils;
import easik.ui.OptionsDialog;
import easik.ui.Option;

import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JPasswordField;
import javax.swing.JCheckBox;
import easik.ui.SketchFrame;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;

public class DatabaseOptions extends OptionsDialog {

	// The SQL dialect we're using
	private String _dialect;

	/** Thid dialog's parent frame */
	private SketchFrame _theFrame;

	private boolean _skipConnSettings = false;

	/** If available, dialog uses this driver's options as field defaults */
	private DatabaseDriver _dbd;

	/*
	 * Various JThings containing entered information.  Note that not all of
	 * these are used for every database type.
	 */
	private JTextField _hostname, _port, _database, _schema, _username;
	private JPasswordField _password;
	private JCheckBox _quoteIdentifiers;

	/**
	 * Creates and displays a new modal database options dialog.
	 *
	 * @param dialect the SQL driver type, such as "MySQL" or "PostgreSQL"
	 * @param dbd an existing driver whos settings will be used as the defualt for the current dialog
	 * @param sketchFrame the SketchFrame to attach this modal dialog box to
	 */
	public DatabaseOptions(String dialect, DatabaseDriver dbd, SketchFrame sketchFrame) {
		super(sketchFrame, "Database connection options");
		_dialect = dialect;
		_theFrame = sketchFrame;
		setSize(425, 350);
		_dbd = dbd;
		showDialog();
	}

	/**
	 * Creates and displays a new modal database options dialog, but doesn't
	 * ask for username/password/hostname/port, but only database name (and
	 * related).  This is used when getting a driver for a non-connection (for
	 * instance, for an SQL text export).
	 * @param dialect the SQL driver type, such as "MySQL" or "PostgreSQL"
	 * @param dbd an existing driver whos settings will be used as the defualt for the current dialog
	 * @param sketchFrame the SketchFrame to attach this modal dialog box to
	 */
	public DatabaseOptions(String dialect, DatabaseDriver dbd, SketchFrame sketchFrame, boolean noConnection) {
		super(sketchFrame, "Database options");
		_dialect = dialect;
		_theFrame = sketchFrame;
		setSize(425, noConnection ? 250 : 350);
		_skipConnSettings = noConnection;
		_dbd = dbd;
		showDialog();
	}
	/**
	 * Creates and displays a new modal database options dialog.
	 *
	 * @param driver the SQL driver type, such as "MySQL" or "PostgreSQL"
	 * @param sketchFrame the SketchFrame to attach this modal dialog box to
	 */
	public DatabaseOptions(String dialect, SketchFrame sketchFrame) {
		super(sketchFrame, "Database connection options");
		_dialect = dialect;
		_theFrame = sketchFrame;
		setSize(425, 350);
		showDialog();
	}

	/**
	 * Creates and displays a new modal database options dialog, but doesn't
	 * ask for username/password/hostname/port, but only database name (and
	 * related).  This is used when getting a driver for a non-connection (for
	 * instance, for an SQL text export).
	 */
	public DatabaseOptions(String dialect, SketchFrame sketchFrame, boolean noConnection) {
		super(sketchFrame, "Database options");
		_dialect = dialect;
		_theFrame = sketchFrame;
		setSize(425, noConnection ? 250 : 350);
		_skipConnSettings = noConnection;
		showDialog();
	}

	public List<Option> getOptions() {
		LinkedList<Option> opts = new LinkedList<Option>();

		Map<String, String> saved = _theFrame.getSketch().getConnectionParams();
		if (!_skipConnSettings) {

			String username = saved.get("username"), hostname = saved.get("hostname"), port = saved.get("port");
			if (hostname == null) hostname = "localhost"; // A reasonable hostname default
			opts.add(new Option(new JLabel("Username"), _username = JUtils.textField(username)));
			opts.add(new Option(new JLabel("Password"), _password = (JPasswordField) JUtils.fixHeight(new JPasswordField())));

			opts.add(new Option(new JLabel("Database hostname"), _hostname = JUtils.textField(hostname)));
			opts.add(new Option(new JLabel("Database port"), _port = JUtils.textField(port)));
		}

		String defaultDBName = saved.get("database");
		if (defaultDBName == null) defaultDBName =
			_theFrame.getSketch().getDocInfo().getName().replaceAll("\\W+", "_").replaceFirst("^_+", "").replaceFirst("_+$", "");

		opts.add(new Option(new JLabel("Database name"), _database = JUtils.textField(defaultDBName)));

		if (_dialect.equals("PostgreSQL")) {
			String schema = saved.get("schema");
			if (schema == null) schema = "";
			opts.add(new Option(new JLabel("Schema name"), _schema = JUtils.textField(schema)));
		}

		boolean quoting;
		String quoteSetting = saved.get("quoteIdentifiers");
		if (quoteSetting == null)
			quoteSetting = Easik.getInstance().getSettings().getProperty("sql_quoting");

		quoting = quoteSetting != null && quoteSetting.equals("true");

		_quoteIdentifiers = new JCheckBox("Quote table and column names");
		_quoteIdentifiers.setSelected(quoting);

		opts.add(new Option(new JLabel("Identifier quoting"), _quoteIdentifiers));

		// FIXME -- we should add some help (mouseover? help button?) to describe these in more detail
		if(_dbd != null) {
			// If we have an existing connection, reset everything to its current settings:
			if(_dbd.hasOption("hostname"))
				_hostname.setText((String)_dbd.getOption("hostname"));
			if(_dbd.hasOption("port"))
				_port.setText((String)_dbd.getOption("port"));
			if(_dbd.hasOption("username"))
				_username.setText((String)_dbd.getOption("username"));
			if(_dbd.hasOption("database"))
				_database.setText((String)_dbd.getOption("database"));
			if(_dialect.equals("PostgreSQL") && _dbd.hasOption("schema"))
				_schema.setText((String)_dbd.getOption("schema"));
			if(_dbd.hasOption("quoteIdentifiers"))
				_quoteIdentifiers.setSelected(Boolean.parseBoolean((String)_dbd.getOption("quoteIdentifiers")));
		}

		return opts;
	}

	public Map<String, String> getParams() {
		HashMap<String, String> options = new HashMap<String, String>();

		if (!_skipConnSettings) {
			options.put("hostname", _hostname.getText());
			options.put("port", _port.getText());
			options.put("username", _username.getText());
			options.put("password", new String(_password.getPassword()));
		}

		options.put("database", _database.getText());
		if (_dialect.equals("PostgreSQL"))
			options.put("schema", _schema.getText());
		options.put("quoteIdentifiers", _quoteIdentifiers.isSelected() ? "true" : "false");

		// We don't prompt for the pk/fk settings currently, but include them anyway.
		String pkFormat = Easik.getInstance().getSettings().getProperty("sql_pk_columns");
		String fkFormat = Easik.getInstance().getSettings().getProperty("sql_fk_columns");
		if (pkFormat != null)
			options.put("pkFormat", pkFormat);
		if (fkFormat != null)
			options.put("fkFormat", fkFormat);

		return options;
	}

	public boolean verify() {
		// FIXME -- make sure they entered at least a database name
		// Perhaps some other data sanity checks?
		// - numeric port
		// - hostname consisting only of letters, numbers, and .'s.
		
		return true;
	}
}
