package easik.ui.menu.popup;

import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.JOptionPane;

import easik.ui.SketchFrame;
import easik.sketch.Sketch;
import easik.sketch.constraint.Constraint;
import easik.sketch.edge.SketchEdge;
import easik.sketch.vertex.EntityNode;


/**
 * Action used to delete a node(s) from a sketch.
 */
public class DeleteFromSketchAction extends AbstractAction {
	
	/** The sketch frame in which the deleting occurs */
	SketchFrame _theFrame;

	/**
	 * The delete button gets initialized
	 */
	public DeleteFromSketchAction(SketchFrame inFrame) {
		super("Delete");
		_theFrame = inFrame;
		putValue(
			AbstractAction.SHORT_DESCRIPTION,
			"Delete selection.");
	}

	/**
	 * When the action is performed, selection is deleted if possible.  Error is displayed 
	 * if no graph item is selected.
	 * 
	 * @param e The action event
	 */
	public void actionPerformed(ActionEvent e) {
		Sketch _ourSketch = _theFrame.getSketch();

		// The confirm delete message. If we're currently synced with a database, add that to the message;
		String confirm = _ourSketch.isSynced() ? "Warning: this sketch is currently synced with a database; delete and break synchronization?"
			: "Are you sure you want to delete selected item(s)?";

		if(JOptionPane.showConfirmDialog(_theFrame, 
					confirm, 
					"Warning!", 
					JOptionPane.OK_CANCEL_OPTION,
					JOptionPane.WARNING_MESSAGE) == JOptionPane.CANCEL_OPTION){
			return;
		}
		Object[] currentSelection = _ourSketch.getSelectionCells();

		if (currentSelection.length == 0) {
			JOptionPane.showMessageDialog(
					_theFrame,
					"Operation must be performed with something selected",
					"Error",
					JOptionPane.ERROR_MESSAGE);
		} else {
			_ourSketch.getSketchModel().beginUpdate();

			// Delete constraints, edges, and entities in that order,
			// to avoid deletion of an entity causing deletion of edges or constraints
			// and thereby causing a problem trying to remove something that is
			// already removed.

			// First, delete any constraints:
			for (Object o: currentSelection) {
				if (o instanceof Constraint) {
					_ourSketch.removeConstraint((Constraint) o);
				}
			}
			// THEN any edges:
			for (Object o: currentSelection) {
				if (o instanceof SketchEdge) {
					_ourSketch.removeEdge((SketchEdge) o);
				}
			}

			// Then finally, any entities.
			for (Object o: currentSelection) {
				if (o instanceof EntityNode) {
					_ourSketch.removeEntity((EntityNode) o);
				}
			}
			_ourSketch.setDirty();
			_ourSketch.getSketchModel().endUpdate();
			_ourSketch.setSynced(false);
		}

		// Clear selection after things have been deleted		
		_ourSketch.clearSelection();
	}
}
