package easik.ui.tree;

import java.util.ArrayList;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import javax.swing.undo.AbstractUndoableEdit;

import easik.graph.EasikGraphModel;
import easik.sketch.attribute.EntityAttribute;
import easik.sketch.keys.UniqueKey;
import easik.sketch.constraint.*;
import easik.sketch.path.SketchPath;
import easik.sketch.vertex.EntityNode;
import easik.ui.SketchFrame;
import easik.ui.menu.AddCommutativeDiagramAction;
import easik.ui.menu.AddProductConstraintAction;
import easik.ui.menu.AddPullbackConstraintAction;
import easik.ui.menu.AddSumConstraintAction;
import easik.ui.menu.popup.NewEntityAction;
import easik.ui.tree.popup.AddAttributeAction;
import easik.ui.tree.popup.AddPathAction;
import easik.ui.tree.popup.AddUniqueKeyAction;
import easik.ui.tree.popup.DeleteAttributeAction;
import easik.ui.tree.popup.DeleteConstraintAction;
import easik.ui.tree.popup.DeleteEntityAction;
import easik.ui.tree.popup.DeleteUniqueKeyAction;
import easik.ui.tree.popup.EditAttributeAction;
import easik.ui.tree.popup.EditUniqueKeyAction;
import easik.ui.tree.popup.DeletePathAction;
import easik.ui.tree.popup.RenameEntityAction;


/**
 * Class to display the information tree.  This tree contains information about entities, attributes 
 * and constraints.
 * 
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @version 2006-08-02 Kevin Green
 */
public class SketchInfoTreeUI extends JScrollPane {

	/** The information tree */
	private JTree _infoTree;

	/** The information tree model */
	private DefaultTreeModel _infoTreeModel;

	/** The top node of the tree */
	private DefaultMutableTreeNode _topNode;

	/** The entities tree node */
	private DefaultMutableTreeNode _tree_entities;

	/** The constraints tree node */
	private DefaultMutableTreeNode _tree_constraints;

	/** The commutative diagram tree node */
	private DefaultMutableTreeNode _tree_constraints_commutative;

	/** The product constraint tree node */
	private DefaultMutableTreeNode _tree_constraints_product;

	/** The pullback constraint tree node */
	private DefaultMutableTreeNode _tree_constraints_pullback;

	/** The equalizer constraint tree node */
	private DefaultMutableTreeNode _tree_constraints_equalizer;

	/** The sum constraint tree node */
	private DefaultMutableTreeNode _tree_constraints_sum;

	/** The popup menu for the tree */
	private JPopupMenu _popupMenu;

	/** The position to place new nodes */
	private Point _newPosition;

	/** Tree popup menu item */
	private JMenuItem _addAttributeItem;

	/** Tree popup menu item */
	private JMenuItem _editAttributeItem;

	/** Tree popup menu item */
	private JMenuItem _deleteAttributeItem;

	/** Tree popup menu item */
	private JMenuItem _addUniqueKeyItem;

	/** Tree popup menu item */
	private JMenuItem _editUniqueKeyItem;

	/** Tree popup menu item */
	private JMenuItem _deleteUniqueKeyItem;

	/** Tree popup menu item */
	private JMenuItem _deleteConstraintItem;

	/** Tree popup menu item */
	private JMenuItem _addEntityItem;

	/** Tree popup menu item */
	private JMenuItem _renameEntityItem;

	/** Tree popup menu item */
	private JMenuItem _deleteEntityItem;

	/** Tree popup menu item */
	private JMenuItem _addCommutativeItem;

	/** Tree popup menu item */
	private JMenuItem _addProductItem;

	/** Tree popup menu item */
	private JMenuItem _addPullbackItem;

	/** Tree popup menu item */
	private JMenuItem _addEqualizerItem;

	/** Tree popup menu item */
	private JMenuItem _addSumItem;

	/** Tree popup menu item */
	private JMenuItem _addPathItem;

	/** Tree popup menu item */
	private JMenuItem _deletePathItem;

	/** The frame in which this tree exists */
	private SketchFrame _theFrame;
	
	/** 
	 * An ArrayList of Strings which represent an expansion state of the info tree. The tree
	 * can then be reverted back to the state as defined in expanState.
	 */
	private ArrayList<String> expanState;

	/**
	 * Default constructor
	 */
	public SketchInfoTreeUI(SketchFrame inFrame){

		//Create the top node.
		_topNode = new DefaultMutableTreeNode("EA Sketch");

		_theFrame = inFrame;
		_newPosition = new Point(50, 50);

		//Create a tree that allows one selection at a time.
		_infoTreeModel = new DefaultTreeModel(_topNode);
		_infoTree = new JTree(_infoTreeModel);
		_infoTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		_infoTree.setRootVisible(false); //Hides root node
		_infoTree.setShowsRootHandles(true); //Shows nodes off root

		DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
		renderer.setOpenIcon(null);
		renderer.setClosedIcon(null);
		renderer.setLeafIcon(null);
		_infoTree.setCellRenderer(renderer);

		//Add tree into scroll pane and then to this pane
		this.setViewportView(_infoTree);
		this.setMinimumSize(new Dimension(150, 200));

		_popupMenu = new JPopupMenu();
		buildPopupMenu();

		//Create Category Heads
		_tree_entities = new DefaultMutableTreeNode("Entities");
		_topNode.add(_tree_entities);

		_tree_constraints = new DefaultMutableTreeNode("Constraints");
		_topNode.add(_tree_constraints);

		_tree_constraints_commutative = new DefaultMutableTreeNode("Commutative Diagrams");
		_tree_constraints.add(_tree_constraints_commutative);

		_tree_constraints_product = new DefaultMutableTreeNode("Product Constraints");
		_tree_constraints.add(_tree_constraints_product);

		_tree_constraints_pullback = new DefaultMutableTreeNode("Pullback Constraints");
		_tree_constraints.add(_tree_constraints_pullback);

		_tree_constraints_equalizer = new DefaultMutableTreeNode("Equalizer Constraints");
		_tree_constraints.add(_tree_constraints_equalizer);

		_tree_constraints_sum = new DefaultMutableTreeNode("Sum Constraints");
		_tree_constraints.add(_tree_constraints_sum);

		//Initialize ArrayList of expansion state information
		expanState = new ArrayList<String>();
	}
	
	/**
	 * Creates the popup menu
	 */
	private void buildPopupMenu() {
		_popupMenu.add(_addAttributeItem = new JMenuItem(new AddAttributeAction(_theFrame)));
		_popupMenu.add(_editAttributeItem = new JMenuItem(new EditAttributeAction(_theFrame)));
		_popupMenu.add(_deleteAttributeItem = new JMenuItem(new DeleteAttributeAction(_theFrame)));
		_popupMenu.add(_addUniqueKeyItem = new JMenuItem(new AddUniqueKeyAction(_theFrame)));
		_popupMenu.add(_editUniqueKeyItem = new JMenuItem(new EditUniqueKeyAction(_theFrame)));
		_popupMenu.add(_deleteUniqueKeyItem = new JMenuItem(new DeleteUniqueKeyAction(_theFrame)));
		_popupMenu.add(_deleteConstraintItem = new JMenuItem(new DeleteConstraintAction(_theFrame)));
		_popupMenu.add(_addEntityItem = new JMenuItem(new NewEntityAction(_newPosition, _theFrame)));
		_popupMenu.add(_renameEntityItem = new JMenuItem(new RenameEntityAction(_theFrame)));
		_popupMenu.add(_deleteEntityItem = new JMenuItem(new DeleteEntityAction(_theFrame)));
		_popupMenu.add(_addCommutativeItem = new JMenuItem(new AddCommutativeDiagramAction(_theFrame)));
		_popupMenu.add(_addProductItem = new JMenuItem(new AddProductConstraintAction(_theFrame)));
		_popupMenu.add(_addPullbackItem = new JMenuItem(new AddPullbackConstraintAction(_theFrame)));
		_popupMenu.add(_addSumItem = new JMenuItem(new AddSumConstraintAction(_theFrame)));
		_popupMenu.add(_addPathItem = new JMenuItem(new AddPathAction(_theFrame)));
		_popupMenu.add(_deletePathItem = new JMenuItem(new DeletePathAction(_theFrame)));
		
		_infoTree.addMouseListener(new MouseAdapter() {

			public void mousePressed(MouseEvent evt) {
				mouseReleased(evt);
			}
			public void mouseReleased(MouseEvent evt) {
				if (evt.isPopupTrigger()) {
					_infoTree.setSelectionRow(_infoTree.getClosestRowForLocation(evt.getX(), evt.getY()));
					if(setPopMenuItems()){
						storeExpansion();
						_newPosition.setLocation(_theFrame.getSketch().getNewPosition(10));
						_popupMenu.show(evt.getComponent(), evt.getX(), evt.getY());
					}
				}
			}
		});
	}

	/**
	 * Stores the current expansion state of the info tree.
	 */
	public void storeExpansion(){
		expanState.clear();
		for(int i=0; i < _infoTree.getRowCount(); i++){
			expanState.add(TreeExpansionUtil.getExpansionState(_infoTree, i));
		}
	}

	/**
	 * Reverts the expansion state of the info tree to the last
	 * stored state.
	 */
	public void revertExpansion(){
		for(int i=0; i<expanState.size(); i++){
			TreeExpansionUtil.restoreExpansionState(_infoTree, i, expanState.get(i));
		}
	}
	
	/**
	 * Sets which of the menu items will be visible
	 * 
	 * @return true if the popup should be displayed, false otherwise
	 */
	public boolean setPopMenuItems(){
		//If there is nothing seleceted then just do nothing
		if(_theFrame.getInfoTreeUI().getInfoTree().isSelectionEmpty()){
			return false;
		}
		// If we're in manipulation mode we do not want to be able to edit the sketch
		else if(_theFrame.getMode() == SketchFrame.Mode.MANIPULATE){
			return false;
		}
	
		//Get currently selected object
		DefaultMutableTreeNode curSelected = (DefaultMutableTreeNode) 
					_theFrame.getInfoTreeUI().getInfoTree().getSelectionPath().getLastPathComponent();

		//Hide all elements
		for(Component c : _popupMenu.getComponents()){
			c.setVisible(false);
		}

		//Check what is currently selected
		if (curSelected instanceof EntityNode) {
			_addAttributeItem.setVisible(true);
			_addUniqueKeyItem.setVisible(true);
			_renameEntityItem.setVisible(true);
			_deleteEntityItem.setVisible(true);
		}
		else if (curSelected instanceof EntityAttribute) {
			_editAttributeItem.setVisible(true);
			_deleteAttributeItem.setVisible(true);
		}
		else if (curSelected instanceof UniqueKey) {
			_editUniqueKeyItem.setVisible(true);
			_deleteUniqueKeyItem.setVisible(true);
		}
		else if (curSelected instanceof Constraint) {
			if(curSelected instanceof SumConstraint || curSelected instanceof ProductConstraint || curSelected instanceof CommutativeDiagram)
				_addPathItem.setVisible(true);
			_deleteConstraintItem.setVisible(true);
		}
		else if (curSelected instanceof SketchPath) {
			Object myConst = curSelected.getParent();
			if (myConst instanceof SumConstraint || myConst instanceof ProductConstraint || myConst instanceof CommutativeDiagram) {
				_deletePathItem.setVisible(true);
			}
		}
		else if (curSelected == _tree_entities) {
			_addEntityItem.setVisible(true);
		}
		else if (curSelected == _tree_constraints) {
			_addCommutativeItem.setVisible(true);
			_addProductItem.setVisible(true);
			_addPullbackItem.setVisible(true);
			_addSumItem.setVisible(true);
		}
		else if (curSelected == _tree_constraints_commutative) {
			_addCommutativeItem.setVisible(true);
		}
		else if (curSelected == _tree_constraints_product) {
			_addProductItem.setVisible(true);
		}
		else if (curSelected == _tree_constraints_pullback) {
			_addPullbackItem.setVisible(true);
		}
		else if (curSelected == _tree_constraints_equalizer) {
			_addEqualizerItem.setVisible(true);
		}
		else if (curSelected == _tree_constraints_sum) {
			_addSumItem.setVisible(true);
		}
		else{
			return false;
		}
		return true;
	}
	
	/**
	 * Refreshes the tree so updates visualize properly
	 */
	public void refreshTree(){
		storeExpansion();
		_infoTreeModel.reload(_topNode);
		revertExpansion();
	}
	
	/**
	 * Refreshes the tree so updates visualize properly
	 * 
	 * @param inNode The node to be refreshed
	 */
	public void refreshTree(DefaultMutableTreeNode inNode){
		storeExpansion();
		_infoTreeModel.reload(inNode);
		revertExpansion();
	}
	
	/**
	 * Returns the information tree model
	 * 
	 * @return The information tree model
	 */
	public DefaultTreeModel getInfoTreeModel(){
		return _infoTreeModel;
	}
	
	/**
	 * Returns the information tree
	 * 
	 * @return The information tree
	 */
	public JTree getInfoTree(){
		return _infoTree;
	}
	
	/**
	 * Returns the Entities tree node
	 * 
	 * @return The Entities tree node
	 */
	public DefaultMutableTreeNode getEntities(){
		return _tree_entities;
	}
	
	/**
	 * Returns the Contraints tree node
	 * 
	 * @return The Contraints tree node
	 */
	public DefaultMutableTreeNode getContraints(){
		return _tree_constraints;
	}
	
	/**
	 * Adds an entity and its attributes to the tree
	 * 
	 * @param inEntity The entity to be added to the tree
	 */
	public void addEntity(final EntityNode entity) {
		EasikGraphModel model = (EasikGraphModel) _theFrame.getSketch().getModel();
		model.beginUpdate();

		_addEntity(entity);
		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_removeEntity(entity);
			}
			public void redo() {
				super.redo();
				_addEntity(entity);
			}
		});

		model.endUpdate();
	}

	/**
	 * Internal method to actually add an entity from the tree.  Separated
	 * because of undo/redo support.
	 */
	private void _addEntity(EntityNode entity) {
		_tree_entities.add(entity.getTreeNode());

		refreshTree(_tree_entities);
		_infoTree.scrollPathToVisible(new TreePath(entity.getTreeNode().getPath()));
	}
	
	/**
	 * Removes an entity from the tree.  Undoable.
	 * 
	 * @param toRemove Entity to be removed from the tree
	 */
	public void removeEntity(final EntityNode toRemove) {
		EasikGraphModel model = (EasikGraphModel) _theFrame.getSketch().getModel();
		model.beginUpdate();

		_removeEntity(toRemove);
		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_addEntity(toRemove);
			}
			public void redo() {
				super.redo();
				_removeEntity(toRemove);
			}
		});

		model.endUpdate();
	}

	/**
	 * Internal method to actually remove an entity from the tree.  Separated
	 * because of undo/redo support.
	 */
	private void _removeEntity(EntityNode toRemove) {
		_tree_entities.remove(toRemove.getTreeNode());

		refreshTree(_tree_entities);
	}

	/**
	 * Add a constraint to the info tree.  This action will factor into the
	 * Sketch's undo/redo history as a single action.
	 * 
	 * @param c The constraint to add
	 * @since 2006-05-30 Vera Ranieri
	 */
	public void addConstraint(final Constraint constraint){
		EasikGraphModel model = (EasikGraphModel) _theFrame.getSketch().getModel();
		model.beginUpdate();

		_addConstraint(constraint);

		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_removeConstraint(constraint);
			}
			public void redo() {
				super.redo();
				_addConstraint(constraint);
			}
		});

		model.endUpdate();

		DefaultMutableTreeNode node = constraint.getTreeNode();
		_infoTree.scrollPathToVisible(new TreePath(node.getPath()));
	}

	/**
	 * Internal method that actually does the work of adding the constraint to
	 * the info tree; this is separate from the public method so that the
	 * public method can handle undo/redo controlling.
	 */
	private void _addConstraint(Constraint constraint) {
		DefaultMutableTreeNode node = constraint.getTreeNode();

		if (constraint instanceof ProductConstraint)
			_tree_constraints_product.add(node);
		else if (constraint instanceof SumConstraint)
			_tree_constraints_sum.add(node);
		else if (constraint instanceof PullbackConstraint)
			_tree_constraints_pullback.add(node);
		else if (constraint instanceof EqualizerConstraint)
			_tree_constraints_equalizer.add(node);
		else if (constraint instanceof CommutativeDiagram)
			_tree_constraints_commutative.add(node);
	
		//Add the paths to the constraint
		for (SketchPath path: constraint.getPaths()) {
			node.add(path);
		}

		refreshTree((DefaultMutableTreeNode) node.getParent());
	}

	/**
	 * Remove a constraint from the info tree.  This action will factor into
	 * the Sketch's undo/redo history as a single action.
	 * 
	 * @param c The constraint to remove
	 * @since 2006-05-30 Vera Ranieri
	 */
	public void removeConstraint(final Constraint constraint) {
		storeExpansion();

		EasikGraphModel model = (EasikGraphModel) _theFrame.getSketch().getModel();
		model.beginUpdate();

		_removeConstraint(constraint);
		model.postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_addConstraint(constraint);
			}
			public void redo() {
				super.redo();
				_removeConstraint(constraint);
			}
		});

		model.endUpdate();
		revertExpansion();
	}

	/**
	 * Internal method that actually does the work of adding the constraint to
	 * the info tree; this is separate from the public method so that the
	 * public method can handle undo/redo controlling.
	 */
	private void _removeConstraint(Constraint c) {
		DefaultMutableTreeNode constraint = c.getTreeNode();

		if ( c instanceof CommutativeDiagram )
			_tree_constraints_commutative.remove(constraint);
		else if ( c instanceof ProductConstraint )
			_tree_constraints_product.remove(constraint);
		else if ( c instanceof PullbackConstraint )
			_tree_constraints_pullback.remove(constraint);
		else if ( c instanceof EqualizerConstraint )
			_tree_constraints_equalizer.remove(constraint);
		else if ( c instanceof SumConstraint )
			_tree_constraints_sum.remove(constraint);

		refreshTree();
	}
}
